# Test freezing inline-asm code.

# ruff: noqa: F821 - @asm_rv32 decorator adds names to function scope

import micropython


@micropython.asm_rv32
def asm_add(a0, a1):
    add(a0, a0, a1)


@micropython.asm_rv32
def asm_add1(a0) -> object:
    slli(a0, a0, 1)
    addi(a0, a0, 3)


@micropython.asm_rv32
def asm_cast_bool(a0) -> bool:
    pass


@micropython.asm_rv32
def asm_shift_int(a0) -> int:
    slli(a0, a0, 29)


@micropython.asm_rv32
def asm_shift_uint(a0) -> uint:
    slli(a0, a0, 29)



from pyb import CAN
import time
import errno

# Test the various receive IRQs, including overflow

rx_overflow = False

REASONS = ["received", "full", "overflow"]

# CAN IDs
ID_SPAM = 0x345  # messages spammed into the receive FIFO
ID_ACK_OFLOW = 0x055  # message the receiver sends after it's seen an overflow
ID_AFTER = 0x100  # message the sender sends after the ACK


def cb0(bus, reason):
    global rx_overflow
    if reason != 0 and not rx_overflow:
        # exact timing of 'received' callbacks depends on controller type,
        # so only log the other two
        print("rx0 reason", REASONS[reason])
    if reason == 2:
        rx_overflow = True


# Accept all standard IDs on FIFO 0
def _enable_accept_all():
    if hasattr(CAN, "MASK"):  # FD-CAN controller
        can.setfilter(0, CAN.RANGE, 0, (0x0, 0x7FF), extframe=False)
    else:
        can.setfilter(0, CAN.MASK16, 0, (0, 0, 0, 0), extframe=False)


# Receiver
def instance0():
    _enable_accept_all()
    can.rxcallback(0, cb0)

    multitest.next()
    multitest.wait("sender ready")
    multitest.broadcast("receiver ready")

    while not rx_overflow:
        pass  # Resume ASAP after FIFO0 overflows

    can.send(b"overflow", ID_ACK_OFLOW)

    # drain the receive FIFO, making sure we read at least on ID_SPAM message
    rxed_spam = False
    while can.any(0):
        msg = can.recv(0, timeout=0)
        assert msg[0] == ID_SPAM
        rxed_spam = True
    print("rxed_spam", rxed_spam)

    # This should be the one message with ID_AFTER, there may be one or two spam messages as well
    for _ in range(10):
        msg = can.recv(0, timeout=500)
        if msg[0] == ID_AFTER:
            print(msg)
            break

    # RX FIFO should be empty now
    print("any", can.any(0))


# Sender
def instance1():
    _enable_accept_all()
    multitest.next()
    multitest.broadcast("sender ready")
    multitest.wait("receiver ready")

    # Spam out messages until the receiver tells us its RX FIFO is full.
    #
    # The RX FIFO on the receiver can vary from 3 deep (BXCAN) to 25 deep (STM32H7),
    # so we keep sending to it until we see a CAN message on ID_ACK_OFLOW indicating
    # the receiver's FIFO has overflowed
    for i in range(255):
        can.send(bytes([i] * 8), ID_SPAM, timeout=25)
        if can.any(0):
            print(can.recv(0))  # should be ID_ACK_OFLOW
            break
        # on boards like STM32H7 the TX FIFO is really deep, so don't fill it too quickly...
        time.sleep_ms(1)

    # give the receiver some time to make space in the FIFO
    time.sleep_ms(200)

    # send the final message, the receiver should get this one
    can.send(b"aaaaa", ID_AFTER)

    # Sender's RX FIFO should also be empty at this point
    print("any", can.any(0))


can = CAN(1, CAN.NORMAL, baudrate=500_000, sample_point=75)



"""
categories: Modules,json
description: JSON module does not throw exception when object is not serialisable
cause: Unknown
workaround: Unknown
"""

import json

try:
    print(json.dumps(b"shouldn't be able to serialise bytes"))
except TypeError:
    print("TypeError")



"""
categories: Syntax,Operators
description: MicroPython allows := to assign to the iteration variable in nested comprehensions, CPython does not.
cause: MicroPython is optimised for code size. Although it is a syntax error to assign to the iteration variable in a standard comprehension (same as CPython), it doesn't check if an inner nested comprehension assigns to the iteration variable of the outer comprehension.
workaround: Do not use := to assign to the iteration variable of a comprehension.
"""

print([[(j := i) for i in range(2)] for j in range(2)])



# Test that SLEEP_ENx registers are preserved over a call to machine.lightsleep().

import sys
from machine import mem32, lightsleep
import unittest

is_rp2350 = "RP2350" in sys.implementation._machine

if is_rp2350:
    CLOCK_BASE = 0x40010000
    SLEEP_EN0 = CLOCK_BASE + 0xB4
    SLEEP_EN1 = CLOCK_BASE + 0xB8
    TO_DISABLE_EN0 = 1 << 30  # SHA256
    TO_DISABLE_EN1 = 1 << 4  # SRAM0
else:
    CLOCK_BASE = 0x40008000
    SLEEP_EN0 = CLOCK_BASE + 0xA8
    SLEEP_EN1 = CLOCK_BASE + 0xAC
    TO_DISABLE_EN0 = 1 << 28  # SRAM0
    TO_DISABLE_EN1 = 1 << 0  # SRAM4


class Test(unittest.TestCase):
    def setUp(self):
        self.orig_sleep_en0 = mem32[SLEEP_EN0]
        self.orig_sleep_en1 = mem32[SLEEP_EN1]

    def tearDown(self):
        mem32[SLEEP_EN0] = self.orig_sleep_en0
        mem32[SLEEP_EN1] = self.orig_sleep_en1

    def test_sleep_en_regs(self):
        print()

        # Disable some bits so the registers aren't just 0xffff.
        mem32[SLEEP_EN0] &= ~TO_DISABLE_EN0
        mem32[SLEEP_EN1] &= ~TO_DISABLE_EN1

        # Get the registers before the lightsleep.
        sleep_en0_before = mem32[SLEEP_EN0] & 0xFFFFFFFF
        sleep_en1_before = mem32[SLEEP_EN1] & 0xFFFFFFFF
        print(hex(sleep_en0_before), hex(sleep_en1_before))

        # Do a lightsleep.
        lightsleep(100)

        # Get the registers after a lightsleep.
        sleep_en0_after = mem32[SLEEP_EN0] & 0xFFFFFFFF
        sleep_en1_after = mem32[SLEEP_EN1] & 0xFFFFFFFF
        print(hex(sleep_en0_after), hex(sleep_en1_after))

        # Check the registers have not changed.
        self.assertEqual(sleep_en0_before, sleep_en0_after)
        self.assertEqual(sleep_en1_before, sleep_en1_after)


if __name__ == "__main__":
    unittest.main()



# Test the marshal module, stressing edge cases.

try:
    import marshal

    (lambda: 0).__code__
except (AttributeError, ImportError):
    print("SKIP")
    raise SystemExit

ftype = type(lambda: 0)

# Test a large function.


def large_function(arg0, arg1, arg2, arg3):
    # Arguments.
    print(arg0, arg1, arg2, arg3)

    # Positive medium-sized integer (still a small-int though).
    print(1234)

    # Negative small-ish integer.
    print(-20)

    # More than 64 constant objects.
    x = (0,)
    x = (1,)
    x = (2,)
    x = (3,)
    x = (4,)
    x = (5,)
    x = (6,)
    x = (7,)
    x = (8,)
    x = (9,)
    x = (10,)
    x = (11,)
    x = (12,)
    x = (13,)
    x = (14,)
    x = (15,)
    x = (16,)
    x = (17,)
    x = (18,)
    x = (19,)
    x = (20,)
    x = (21,)
    x = (22,)
    x = (23,)
    x = (24,)
    x = (25,)
    x = (26,)
    x = (27,)
    x = (28,)
    x = (29,)
    x = (30,)
    x = (31,)
    x = (32,)
    x = (33,)
    x = (34,)
    x = (35,)
    x = (36,)
    x = (37,)
    x = (38,)
    x = (39,)
    x = (40,)
    x = (41,)
    x = (42,)
    x = (43,)
    x = (44,)
    x = (45,)
    x = (46,)
    x = (47,)
    x = (48,)
    x = (49,)
    x = (50,)
    x = (51,)
    x = (52,)
    x = (53,)
    x = (54,)
    x = (55,)
    x = (56,)
    x = (57,)
    x = (58,)
    x = (59,)
    x = (60,)
    x = (61,)
    x = (62,)
    x = (63,)
    x = (64,)

    # Small jump.
    x = 0
    while x < 2:
        print("loop", x)
        x += 1

    # Large jump.
    x = 0
    while x < 2:
        try:
            try:
                try:
                    print
                except Exception as e:
                    print
                finally:
                    print
            except Exception as e:
                print
            finally:
                print
        except Exception as e:
            print
        finally:
            print("loop", x)
        x += 1


code = marshal.dumps(large_function.__code__)
ftype(marshal.loads(code), {"print": print})(0, 1, 2, 3)



print("foobar".startswith("foo"))
print("foobar".startswith("Foo"))
print("foobar".startswith("foo1"))
print("foobar".startswith("foobar"))
print("foobar".startswith(""))

print("1foobar".startswith("foo", 1))
print("1foo".startswith("foo", 1))
print("1foo".startswith("1foo", 1))
print("1fo".startswith("foo", 1))
print("1fo".startswith("foo", 10))

print("1foobar".startswith("foo", 1, 5))
print("1foobar".startswith("foo", 1, 4))
print("1foobar".startswith("foo", 1, 3))
print("1foobar".startswith("oo", 2, 4))
print("1foobar".startswith("o", 3, 4))
print("1foobar".startswith("o", 4, 4))
print("1foobar".startswith("o", 5, 4))

print("foobar".startswith("foo", None, None))
print("foobar".startswith("foo", None, 3))
print("foobar".startswith("foo", None, 2))
print("foobar".startswith("bar", 3, None))


print("foobar".startswith(("foo", "sth")))
print("foobar".startswith(("sth", "foo")))
print("foobar".startswith(("sth", "foo2")))
print("foobar".startswith(("foo", )))

try:
    "foobar".startswith(1)
except TypeError:
    print("TypeError")



# test load/store opcodes


@micropython.asm_rv32
def l():
    li(a5, 4)
    addi(sp, sp, -12)
    li(a0, 0x123)
    c_swsp(a0, 0)
    addi(a1, a0, 0x111)
    c_swsp(a1, 4)
    addi(a2, a1, 0x111)
    c_swsp(a2, 8)
    mv(a4, sp)
    c_lw(a3, 0(a4))
    bne(a3, a0, END)
    addi(a5, a5, -1)
    lw(a3, 4(a4))
    bne(a3, a1, END)
    addi(a5, a5, -1)
    lhu(a3, 8(a4))
    bne(a3, a2, END)
    addi(a5, a5, -1)
    lbu(a0, 8(a4))
    addi(a0, a0, 0x300)
    bne(a0, a2, END)
    addi(a5, a5, -1)
    label(END)
    addi(sp, sp, 12)
    mv(a0, a5)


print(l())


@micropython.asm_rv32
def s():
    li(a5, 4)
    addi(sp, sp, -12)
    c_swsp(zero, 0)
    c_swsp(zero, 4)
    c_swsp(zero, 8)
    li(a0, 0x12345)
    mv(a4, sp)
    c_sw(a0, 0(a4))
    sh(a0, 4(a4))
    sb(a0, 8(a4))
    li(a1, 0xFFFF)
    and_(a1, a0, a1)
    andi(a2, a0, 0xFF)
    lw(a3, 0(sp))
    bne(a3, a0, END)
    addi(a5, a5, -1)
    lw(a3, 4(sp))
    bne(a3, a1, END)
    addi(a5, a5, -1)
    lw(a3, 8(sp))
    bne(a3, a2, END)
    addi(a5, a5, -1)
    label(END)
    addi(sp, sp, 12)
    mv(a0, a5)


print(s())


@micropython.asm_rv32
def lu():
    li(a5, 4)
    addi(sp, sp, -8)
    li(a0, 0xF1234567)
    c_swsp(a0, 0)
    c_swsp(a0, 4)
    lh(a1, 0(sp))
    blt(a1, zero, END)
    addi(a5, a5, -1)
    lb(a2, 4(sp))
    blt(a2, zero, END)
    addi(a5, a5, -1)
    label(END)
    addi(sp, sp, 8)
    mv(a0, a5)


print(lu())



# Test VfsRom filesystem.

try:
    import errno, sys, struct, os, uctypes, vfs

    vfs.VfsRom
except (ImportError, AttributeError):
    print("SKIP")
    raise SystemExit

try:
    import select
except ImportError:
    select = None

import unittest

IFDIR = 0x4000
IFREG = 0x8000

SEEK_SET = 0
SEEK_CUR = 1
SEEK_END = 2

# An mpy file with four constant objects: str, bytes, long-int, float.
test_mpy = (
    # header
    b"M\x06\x00\x1f"  # mpy file header
    b"\x06"  # n_qstr
    b"\x05"  # n_obj
    # qstrs
    b"\x0etest.py\x00"  # qstr0 = "test.py"
    b"\x0f"  # qstr1 = "<module>"
    b"\x0estr_obj\x00"  # qstr2 = "str_obj"
    b"\x12bytes_obj\x00"  # qstr3 = "bytes_obj"
    b"\x0eint_obj\x00"  # qstr4 = "int_obj"
    b"\x12float_obj\x00"  # qstr5 = "float_obj"
    # objects
    b"\x05\x14this is a str object\x00"
    b"\x06\x16this is a bytes object\x00"
    b"\x07\x0a1234567890"  # long-int object
    b"\x08\x041.23"  # float object
    b"\x05\x07str_obj\x00"  # str object of existing qstr
    # bytecode
    b"\x81\x28"  # 21 bytes, no children, bytecode
    b"\x00\x02"  # prelude
    b"\x01"  # simple name (<module>)
    b"\x23\x00"  # LOAD_CONST_OBJ(0)
    b"\x16\x02"  # STORE_NAME(str_obj)
    b"\x23\x01"  # LOAD_CONST_OBJ(1)
    b"\x16\x03"  # STORE_NAME(bytes_obj)
    b"\x23\x02"  # LOAD_CONST_OBJ(2)
    b"\x16\x04"  # STORE_NAME(int_obj)
    b"\x23\x03"  # LOAD_CONST_OBJ(3)
    b"\x16\x05"  # STORE_NAME(float_obj)
    b"\x51"  # LOAD_CONST_NONE
    b"\x63"  # RETURN_VALUE
)


class VfsRomWriter:
    ROMFS_HEADER = b"\xd2\xcd\x31"

    ROMFS_RECORD_KIND_UNUSED = 0
    ROMFS_RECORD_KIND_PADDING = 1
    ROMFS_RECORD_KIND_DATA_VERBATIM = 2
    ROMFS_RECORD_KIND_DATA_POINTER = 3
    ROMFS_RECORD_KIND_DIRECTORY = 4
    ROMFS_RECORD_KIND_FILE = 5

    def __init__(self):
        self._dir_stack = [(None, bytearray())]

    def _encode_uint(self, value):
        encoded = [value & 0x7F]
        value >>= 7
        while value != 0:
            encoded.insert(0, 0x80 | (value & 0x7F))
            value >>= 7
        return bytes(encoded)

    def _pack(self, kind, payload):
        return self._encode_uint(kind) + self._encode_uint(len(payload)) + payload

    def _extend(self, data):
        buf = self._dir_stack[-1][1]
        buf.extend(data)
        return len(buf)

    def finalise(self):
        _, data = self._dir_stack.pop()
        encoded_kind = VfsRomWriter.ROMFS_HEADER
        encoded_len = self._encode_uint(len(data))
        if (len(encoded_kind) + len(encoded_len) + len(data)) % 2 == 1:
            encoded_len = b"\x80" + encoded_len
        data = encoded_kind + encoded_len + data
        return data

    def opendir(self, dirname):
        self._dir_stack.append((dirname, bytearray()))

    def closedir(self):
        dirname, dirdata = self._dir_stack.pop()
        dirdata = self._encode_uint(len(dirname)) + bytes(dirname, "ascii") + dirdata
        self._extend(self._pack(VfsRomWriter.ROMFS_RECORD_KIND_DIRECTORY, dirdata))

    def mkdata(self, data):
        assert len(self._dir_stack) == 1
        return self._extend(self._pack(VfsRomWriter.ROMFS_RECORD_KIND_DATA_VERBATIM, data)) - len(
            data
        )

    def mkfile(self, filename, filedata, extra_payload=b""):
        filename = bytes(filename, "ascii")
        payload = self._encode_uint(len(filename))
        payload += filename
        payload += extra_payload
        if isinstance(filedata, tuple):
            sub_payload = self._encode_uint(filedata[0])
            sub_payload += self._encode_uint(filedata[1])
            payload += self._pack(VfsRomWriter.ROMFS_RECORD_KIND_DATA_POINTER, sub_payload)
        else:
            payload += self._pack(VfsRomWriter.ROMFS_RECORD_KIND_DATA_VERBATIM, filedata)
        self._dir_stack[-1][1].extend(self._pack(VfsRomWriter.ROMFS_RECORD_KIND_FILE, payload))


def _make_romfs(fs, files, data_map):
    for filename, contents in files:
        if isinstance(contents, tuple):
            fs.opendir(filename)
            _make_romfs(fs, contents, data_map)
            fs.closedir()
        elif isinstance(contents, int):
            fs.mkfile(filename, data_map[contents])
        else:
            fs.mkfile(filename, contents)


def make_romfs(files, data=None):
    fs = VfsRomWriter()
    data_map = {}
    if data:
        for k, v in data.items():
            data_map[k] = len(v), fs.mkdata(v)
    _make_romfs(fs, files, data_map)
    return fs.finalise()


# A class to test if a value is within a range, needed because MicroPython's range
# doesn't support arbitrary objects.
class Range:
    def __init__(self, lower, upper):
        self._lower = lower
        self._upper = upper

    def __repr__(self):
        return "Range({}, {})".format(self._lower, self._upper)

    def __contains__(self, value):
        return self._lower <= value < self._upper


class TestBase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        fs_inner = make_romfs((("test_inner.txt", b"contents_inner"), ("c.py", b"")))
        cls.romfs = make_romfs(
            (
                ("fs.romfs", 0),
                ("test.txt", b"contents"),
                (
                    "dir",
                    (
                        ("a.py", b"x = 1"),
                        ("b.py", b"x = 2"),
                        ("test.mpy", test_mpy),
                    ),
                ),
            ),
            {0: fs_inner},
        )
        cls.romfs_ilistdir = [
            ("fs.romfs", IFREG, 0, 46),
            ("test.txt", IFREG, 0, 8),
            ("dir", IFDIR, 0, 198),
        ]
        cls.romfs_listdir = [x[0] for x in cls.romfs_ilistdir]
        cls.romfs_listdir_dir = ["a.py", "b.py", "test.mpy"]
        cls.romfs_listdir_bytes = [bytes(x, "ascii") for x in cls.romfs_listdir]
        cls.romfs_addr = uctypes.addressof(cls.romfs)
        cls.romfs_addr_range = Range(cls.romfs_addr, cls.romfs_addr + len(cls.romfs))


class TestEdgeCases(unittest.TestCase):
    def test_empty_romfs(self):
        empty_romfs = make_romfs(())
        self.assertEqual(empty_romfs, bytes([0x80 | ord("R"), 0x80 | ord("M"), ord("1"), 0]))
        fs = vfs.VfsRom(empty_romfs)
        self.assertIsInstance(fs, vfs.VfsRom)
        fs.mount(True, False)
        self.assertEqual(list(fs.ilistdir("")), [])
        self.assertEqual(fs.stat(""), (IFDIR, 0, 0, 0, 0, 0, 0, 0, 0, 0))
        self.assertEqual(fs.statvfs(""), (1, 0, 0, 0, 0, 0, 0, 0, 0, 32767))

    def test_error(self):
        for bad_romfs in (b"", b"xxx", b"not a romfs"):
            with self.assertRaises(OSError) as ctx:
                vfs.VfsRom(bad_romfs)
            self.assertEqual(ctx.exception.errno, errno.ENODEV)

    def test_unknown_record(self):
        fs = VfsRomWriter()
        fs._extend(fs._pack(VfsRomWriter.ROMFS_RECORD_KIND_PADDING, b"payload"))
        fs.mkfile(
            "test",
            b"contents",
            extra_payload=fs._pack(VfsRomWriter.ROMFS_RECORD_KIND_PADDING, b"pad"),
        )
        romfs = fs.finalise()
        fs = vfs.VfsRom(romfs)
        self.assertEqual(list(fs.ilistdir("")), [("test", IFREG, 0, 8)])
        with fs.open("test", "rb") as f:
            self.assertEqual(f.read(), b"contents")


class TestCorrupt(unittest.TestCase):
    def test_corrupt_filesystem(self):
        # Make the filesystem length bigger than the buffer.
        romfs = bytearray(make_romfs(()))
        romfs[3] = 0x01
        with self.assertRaises(OSError):
            vfs.VfsRom(romfs)

        # Corrupt the filesystem length.
        romfs = bytearray(make_romfs(()))
        romfs[3] = 0xFF
        with self.assertRaises(OSError):
            vfs.VfsRom(romfs)

        # Corrupt the contents of the filesystem.
        romfs = bytearray(make_romfs(()))
        romfs[3] = 0x01
        romfs.extend(b"\xff\xff")
        fs = vfs.VfsRom(romfs)
        with self.assertRaises(OSError):
            fs.stat("a")
        self.assertEqual(list(fs.ilistdir("")), [])

    def test_corrupt_file_entry(self):
        romfs = make_romfs((("file", b"data"),))

        # Corrupt the length of filename.
        romfs_corrupt = bytearray(romfs)
        romfs_corrupt[7:] = b"\xff" * (len(romfs) - 7)
        fs = vfs.VfsRom(romfs_corrupt)
        with self.assertRaises(OSError):
            fs.stat("file")
        self.assertEqual(list(fs.ilistdir("")), [])

        # Erase the data record (change it to a padding record).
        romfs_corrupt = bytearray(romfs)
        romfs_corrupt[12] = VfsRomWriter.ROMFS_RECORD_KIND_PADDING
        fs = vfs.VfsRom(romfs_corrupt)
        with self.assertRaises(OSError):
            fs.stat("file")
        self.assertEqual(list(fs.ilistdir("")), [])

        # Corrupt the header of the data record.
        romfs_corrupt = bytearray(romfs)
        romfs_corrupt[12:] = b"\xff" * (len(romfs) - 12)
        fs = vfs.VfsRom(romfs_corrupt)
        with self.assertRaises(OSError):
            fs.stat("file")

        # Corrupt the interior of the data record.
        romfs_corrupt = bytearray(romfs)
        romfs_corrupt[13:] = b"\xff" * (len(romfs) - 13)
        fs = vfs.VfsRom(romfs_corrupt)
        with self.assertRaises(OSError):
            fs.stat("file")

        # Change the data record to an indirect pointer and corrupt the length.
        romfs_corrupt = bytearray(romfs)
        romfs_corrupt[12] = VfsRomWriter.ROMFS_RECORD_KIND_DATA_POINTER
        romfs_corrupt[14:18] = b"\xff\xff\xff\xff"
        fs = vfs.VfsRom(romfs_corrupt)
        with self.assertRaises(OSError):
            fs.stat("file")

        # Change the data record to an indirect pointer and corrupt the offset.
        romfs_corrupt = bytearray(romfs)
        romfs_corrupt[12] = VfsRomWriter.ROMFS_RECORD_KIND_DATA_POINTER
        romfs_corrupt[14:18] = b"\x00\xff\xff\xff"
        fs = vfs.VfsRom(romfs_corrupt)
        with self.assertRaises(OSError):
            fs.stat("file")


class TestStandalone(TestBase):
    def test_constructor(self):
        self.assertIsInstance(vfs.VfsRom(self.romfs), vfs.VfsRom)
        with self.assertRaises(TypeError):
            vfs.VfsRom(self.romfs_addr)

    def test_mount(self):
        vfs.VfsRom(self.romfs).mount(True, False)
        with self.assertRaises(OSError):
            vfs.VfsRom(self.romfs).mount(True, True)

    def test_ilistdir(self):
        fs = vfs.VfsRom(self.romfs)
        self.assertEqual(list(fs.ilistdir("")), self.romfs_ilistdir)
        self.assertEqual(list(fs.ilistdir("/")), self.romfs_ilistdir)
        with self.assertRaises(OSError):
            fs.ilistdir("does not exist")

    def test_stat(self):
        fs = vfs.VfsRom(self.romfs)
        self.assertEqual(fs.stat(""), (IFDIR, 0, 0, 0, 0, 0, 289, 0, 0, 0))
        self.assertEqual(fs.stat("/"), (IFDIR, 0, 0, 0, 0, 0, 289, 0, 0, 0))
        self.assertEqual(fs.stat("/test.txt"), (IFREG, 0, 0, 0, 0, 0, 8, 0, 0, 0))
        self.assertEqual(fs.stat("/dir"), (IFDIR, 0, 0, 0, 0, 0, 198, 0, 0, 0))
        with self.assertRaises(OSError):
            fs.stat("/does-not-exist")

    def test_statvfs(self):
        fs = vfs.VfsRom(self.romfs)
        self.assertEqual(fs.statvfs(""), (1, 0, 289, 0, 0, 0, 0, 0, 0, 32767))

    def test_open(self):
        fs = vfs.VfsRom(self.romfs)

        with fs.open("/test.txt", "") as f:
            self.assertEqual(f.read(), "contents")
        with fs.open("/test.txt", "rt") as f:
            self.assertEqual(f.read(), "contents")
        with fs.open("/test.txt", "rb") as f:
            self.assertEqual(f.read(), b"contents")

        with self.assertRaises(OSError) as ctx:
            fs.open("/file-does-not-exist", "")
        self.assertEqual(ctx.exception.errno, errno.ENOENT)

        with self.assertRaises(OSError) as ctx:
            fs.open("/dir", "rb")
        self.assertEqual(ctx.exception.errno, errno.EISDIR)

        with self.assertRaises(OSError):
            fs.open("/test.txt", "w")
        with self.assertRaises(OSError):
            fs.open("/test.txt", "a")
        with self.assertRaises(OSError):
            fs.open("/test.txt", "+")

    def test_file_seek(self):
        fs = vfs.VfsRom(self.romfs)
        with fs.open("/test.txt", "") as f:
            self.assertEqual(f.seek(0, SEEK_SET), 0)
            self.assertEqual(f.seek(3, SEEK_SET), 3)
            self.assertEqual(f.read(), "tents")
            self.assertEqual(f.seek(0, SEEK_SET), 0)
            self.assertEqual(f.seek(100, SEEK_CUR), 8)
            self.assertEqual(f.seek(-1, SEEK_END), 7)
            self.assertEqual(f.read(), "s")
            self.assertEqual(f.seek(1, SEEK_END), 8)
            with self.assertRaises(OSError):
                f.seek(-1, SEEK_SET)
            f.seek(0, SEEK_SET)
            with self.assertRaises(OSError):
                f.seek(-1, SEEK_CUR)
            with self.assertRaises(OSError):
                f.seek(-100, SEEK_END)

    @unittest.skipIf(select is None, "no select module")
    def test_file_ioctl_invalid(self):
        fs = vfs.VfsRom(self.romfs)
        with fs.open("/test.txt", "") as f:
            p = select.poll()
            p.register(f)
            with self.assertRaises(OSError):
                p.poll(0)

    def test_memory_mapping(self):
        fs = vfs.VfsRom(self.romfs)
        with fs.open("/test.txt", "rb") as f:
            addr = uctypes.addressof(f)
            data = memoryview(f)
            self.assertIn(addr, self.romfs_addr_range)
            self.assertIn(addr + len(data), self.romfs_addr_range)
            self.assertEqual(bytes(data), b"contents")


class TestMounted(TestBase):
    def setUp(self):
        self.orig_sys_path = list(sys.path)
        self.orig_cwd = os.getcwd()
        vfs.mount(vfs.VfsRom(self.romfs), "/test_rom")

    def tearDown(self):
        vfs.umount("/test_rom")
        os.chdir(self.orig_cwd)
        sys.path = self.orig_sys_path

    def test_listdir(self):
        self.assertEqual(os.listdir("/test_rom"), self.romfs_listdir)
        self.assertEqual(os.listdir("/test_rom/dir"), self.romfs_listdir_dir)
        self.assertEqual(os.listdir(b"/test_rom"), self.romfs_listdir_bytes)

    def test_chdir(self):
        os.chdir("/test_rom")
        self.assertEqual(os.getcwd(), "/test_rom")
        self.assertEqual(os.listdir(), self.romfs_listdir)

        os.chdir("/test_rom/")
        self.assertEqual(os.getcwd(), "/test_rom")
        self.assertEqual(os.listdir(), self.romfs_listdir)

        # chdir within the romfs is not implemented.
        with self.assertRaises(OSError):
            os.chdir("/test_rom/dir")

    def test_stat(self):
        self.assertEqual(os.stat("/test_rom"), (IFDIR, 0, 0, 0, 0, 0, 289, 0, 0, 0))
        self.assertEqual(os.stat("/test_rom/"), (IFDIR, 0, 0, 0, 0, 0, 289, 0, 0, 0))
        self.assertEqual(os.stat("/test_rom/test.txt"), (IFREG, 0, 0, 0, 0, 0, 8, 0, 0, 0))
        self.assertEqual(os.stat("/test_rom/dir"), (IFDIR, 0, 0, 0, 0, 0, 198, 0, 0, 0))
        with self.assertRaises(OSError):
            os.stat("/test_rom/does-not-exist")

    def test_open(self):
        with open("/test_rom/test.txt") as f:
            self.assertEqual(f.read(), "contents")
        with open("/test_rom/test.txt", "b") as f:
            self.assertEqual(f.read(), b"contents")

        with self.assertRaises(OSError) as ctx:
            open("/test_rom/file-does-not-exist")
        self.assertEqual(ctx.exception.errno, errno.ENOENT)

        with self.assertRaises(OSError) as ctx:
            open("/test_rom/dir")
        self.assertEqual(ctx.exception.errno, errno.EISDIR)

    def test_import_py(self):
        sys.path.append("/test_rom/dir")
        a = __import__("a")
        b = __import__("b")
        self.assertEqual(a.__file__, "/test_rom/dir/a.py")
        self.assertEqual(a.x, 1)
        self.assertEqual(b.__file__, "/test_rom/dir/b.py")
        self.assertEqual(b.x, 2)

    def test_import_mpy(self):
        sys.path.append("/test_rom/dir")
        test = __import__("test")
        self.assertEqual(test.__file__, "/test_rom/dir/test.mpy")
        self.assertEqual(test.str_obj, "this is a str object")
        self.assertEqual(test.bytes_obj, b"this is a bytes object")
        self.assertEqual(test.int_obj, 1234567890)
        self.assertEqual(test.float_obj, 1.23)
        self.assertIn(uctypes.addressof(test.str_obj), self.romfs_addr_range)
        self.assertIn(uctypes.addressof(test.bytes_obj), self.romfs_addr_range)

    def test_romfs_inner(self):
        with open("/test_rom/fs.romfs", "rb") as f:
            romfs_inner = vfs.VfsRom(memoryview(f))

        vfs.mount(romfs_inner, "/test_rom2")

        self.assertEqual(os.listdir("/test_rom2"), ["test_inner.txt", "c.py"])

        sys.path.append("/test_rom2")
        self.assertEqual(__import__("c").__file__, "/test_rom2/c.py")

        with open("/test_rom2/test_inner.txt") as f:
            self.assertEqual(f.read(), "contents_inner")

        with open("/test_rom2/test_inner.txt", "rb") as f:
            addr = uctypes.addressof(f)
            data = memoryview(f)
            self.assertIn(addr, self.romfs_addr_range)
            self.assertIn(addr + len(data), self.romfs_addr_range)

        vfs.umount("/test_rom2")


if __name__ == "__main__":
    unittest.main()



# Test function.__code__ attribute.

try:
    (lambda: 0).__code__
except AttributeError:
    print("SKIP")
    raise SystemExit


def f():
    return a


ftype = type(f)

# Test __code__ access and function constructor.
code = f.__code__
print(type(ftype(code, {})) is ftype)

# Test instantiating multiple code's with different globals dicts.
code = f.__code__
f1 = ftype(code, {"a": 1})
f2 = ftype(code, {"a": 2})
print(f1(), f2())

# Test bad first argument type.
try:
    ftype(None, {})
except TypeError:
    print("TypeError")

# Test bad second argument type.
try:
    ftype(f.__code__, None)
except TypeError:
    print("TypeError")



# test the "data" directive


@micropython.asm_rv32
def ret_num(a0) -> uint:
    slli(a0, a0, 2)
    addi(a0, a0, 16)
    auipc(a1, 0)
    add(a1, a1, a0)
    lw(a0, 0(a1))
    jal(zero, HERE)
    data(4, 0x12345678, 0x20000000, 0x40000000, 0x7FFFFFFF + 1, (1 << 32) - 2)
    label(HERE)


for i in range(5):
    print(hex(ret_num(i)))


@micropython.asm_rv32
def ret_num_la(a0) -> uint:
    slli(a0, a0, 2)
    la(a1, DATA)
    add(a1, a1, a0)
    lw(a0, 0(a1))
    jal(zero, HERE)
    label(DATA)
    data(4, 0x12345678, 0x20000000, 0x40000000, 0x7FFFFFFF + 1, (1 << 32) - 2)
    label(HERE)


for i in range(5):
    print(hex(ret_num_la(i)))



# test that socket.connect() on a non-blocking socket raises EINPROGRESS
# and that an immediate write/send/read/recv does the right thing

import unittest
import errno
import select
import socket
import ssl

# only mbedTLS supports non-blocking mode
ssl_supports_nonblocking = hasattr(ssl, "MBEDTLS_VERSION")


# get the name of an errno error code
def errno_name(er):
    if er == errno.EAGAIN:
        return "EAGAIN"
    if er == errno.EINPROGRESS:
        return "EINPROGRESS"
    return er


# do_connect establishes the socket and wraps it if tls is True.
# If handshake is true, the initial connect (and TLS handshake) is
# allowed to be performed before returning.
def do_connect(self, peer_addr, tls, handshake):
    s = socket.socket()
    s.setblocking(False)
    try:
        print("Connecting to", peer_addr)
        s.connect(peer_addr)
        self.fail()
    except OSError as er:
        print("connect:", errno_name(er.errno))
        self.assertEqual(er.errno, errno.EINPROGRESS)

    # wrap with ssl/tls if desired
    if tls:
        print("wrap socket")
        ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
        s = ssl_context.wrap_socket(s, do_handshake_on_connect=handshake)

    return s


# poll a socket and check the result
def poll(self, s, expect_writable):
    poller = select.poll()
    poller.register(s)
    result = poller.poll(0)
    print("poll:", result)
    if expect_writable:
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0][1], select.POLLOUT)
    else:
        self.assertEqual(result, [])


# do_test runs the test against a specific peer address.
def do_test(self, peer_addr, tls, handshake):
    print()

    # MicroPython plain and TLS sockets have read/write
    hasRW = True

    # MicroPython plain sockets have send/recv
    # MicroPython TLS sockets don't have send/recv
    hasSR = not tls

    # connect + send
    # non-blocking send should raise EAGAIN
    if hasSR:
        s = do_connect(self, peer_addr, tls, handshake)
        poll(self, s, False)
        with self.assertRaises(OSError) as ctx:
            ret = s.send(b"1234")
        print("send error:", errno_name(ctx.exception.errno))
        self.assertEqual(ctx.exception.errno, errno.EAGAIN)
        s.close()

    # connect + write
    # non-blocking write should return None
    if hasRW:
        s = do_connect(self, peer_addr, tls, handshake)
        poll(self, s, tls and handshake)
        ret = s.write(b"1234")
        print("write:", ret)
        if tls and handshake:
            self.assertEqual(ret, 4)
        else:
            self.assertIsNone(ret)
        s.close()

    # connect + recv
    # non-blocking recv should raise EAGAIN
    if hasSR:
        s = do_connect(self, peer_addr, tls, handshake)
        poll(self, s, False)
        with self.assertRaises(OSError) as ctx:
            ret = s.recv(10)
        print("recv error:", errno_name(ctx.exception.errno))
        self.assertEqual(ctx.exception.errno, errno.EAGAIN)
        s.close()

    # connect + read
    # non-blocking read should return None
    if hasRW:
        s = do_connect(self, peer_addr, tls, handshake)
        poll(self, s, tls and handshake)
        ret = s.read(10)
        print("read:", ret)
        self.assertIsNone(ret)
        s.close()


class Test(unittest.TestCase):
    # these tests use a non-existent test IP address, this way the connect takes forever and
    # we can see EAGAIN/None (https://tools.ietf.org/html/rfc5737)
    def test_plain_sockets_to_nowhere(self):
        do_test(self, socket.getaddrinfo("192.0.2.1", 80)[0][-1], False, False)

    @unittest.skipIf(not ssl_supports_nonblocking, "SSL doesn't support non-blocking")
    def test_ssl_sockets_to_nowhere(self):
        do_test(self, socket.getaddrinfo("192.0.2.1", 443)[0][-1], True, False)

    def test_plain_sockets(self):
        do_test(self, socket.getaddrinfo("micropython.org", 80)[0][-1], False, False)

    @unittest.skipIf(not ssl_supports_nonblocking, "SSL doesn't support non-blocking")
    def test_ssl_sockets(self):
        do_test(self, socket.getaddrinfo("micropython.org", 443)[0][-1], True, True)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

import os
import re
import sys

sys.path.insert(0, os.path.join(os.path.dirname(__file__), "../../../tools"))
import boardgen

NUM_PORTS = 16
NUM_PINS_PER_PORT = 8

# Maps pad name to (adc12_periph, adc12_channel).
ADC12_ANA_MAP = {
    "P0_0": (0, 0),
    "P0_1": (0, 1),
    "P0_2": (0, 2),
    "P0_3": (0, 3),
    "P0_4": (0, 4),
    "P0_5": (0, 5),
    "P0_6": (1, 0),
    "P0_7": (1, 1),
    "P1_0": (1, 2),
    "P1_1": (1, 3),
    "P1_2": (1, 4),
    "P1_3": (1, 5),
    "P1_4": (2, 0),
    "P1_5": (2, 1),
    "P1_6": (2, 2),
    "P1_7": (2, 3),
    "P2_0": (2, 4),
    "P2_1": (2, 5),
}


class AlifPin(boardgen.Pin):
    def __init__(self, cpu_pin_name):
        super().__init__(cpu_pin_name)
        self._afs = [("NONE", -1)] * 8

    # Called for each AF defined in the csv file for this pin.
    def add_af(self, af_idx, af_name, af):
        if af == "GPIO":
            self._afs[af_idx] = "GPIO", -1
        elif af.startswith("ANA_S"):
            self._afs[af_idx] = "ANA", int(af[5:])
        else:
            m = re.match(r"([A-Z0-9]+[A-Z])(\d*)_([A-Z0-9_]+)$", af)
            periph, unit, line = m.groups()
            unit = -1 if unit == "" else int(unit)
            self._afs[af_idx] = f"{periph}_{line}", unit

    # Emit the struct which contains the pin instance.
    def definition(self):
        port, pin = self.name()[1:].split("_")
        adc12_periph, adc12_channel = ADC12_ANA_MAP.get(self.name(), (3, 7))
        base = "LPGPIO_BASE" if port == "15" else "GPIO{}_BASE".format(port)
        return (
            "{{\n"
            "    .name = MP_QSTR_P{port}_{pin},\n"
            "    .base = {{ .type = &machine_pin_type }},\n"
            "    .gpio = (GPIO_Type *){base},\n"
            "    .port = PORT_{port},\n"
            "    .pin = PIN_{pin},\n"
            "    .adc12_periph = {adc12_periph},\n"
            "    .adc12_channel = {adc12_channel},\n"
            "    .alt = {{{alt}}},\n"
            "}}".format(
                port=port,
                pin=pin,
                base=base,
                adc12_periph=adc12_periph,
                adc12_channel=adc12_channel,
                alt=", ".join([f"MP_HAL_PIN_ALT({func}, {unit})" for func, unit in self._afs]),
            )
        )

    # Alif cpu names must be "Pn_m".
    @staticmethod
    def validate_cpu_pin_name(cpu_pin_name):
        boardgen.Pin.validate_cpu_pin_name(cpu_pin_name)

        if not (m := re.match("P([0-9]){1,2}_([0-9])", cpu_pin_name)):
            raise boardgen.PinGeneratorError(
                "Invalid cpu pin name '{}', must be 'Pn_m'".format(cpu_pin_name)
            )

        port = int(m.group(1))
        pin = int(m.group(2))
        if not (0 <= port < NUM_PORTS and 0 <= pin < NUM_PINS_PER_PORT):
            raise boardgen.PinGeneratorError("Unknown cpu pin '{}'".format(cpu_pin_name))


class AlifPinGenerator(boardgen.PinGenerator):
    def __init__(self):
        # Use custom pin type above.
        super().__init__(pin_type=AlifPin, enable_af=True)

        # Pre-define the pins (i.e. don't require them to be listed in pins.csv).
        for i in range(NUM_PORTS):
            for j in range(NUM_PINS_PER_PORT):
                self.add_cpu_pin("P{}_{}".format(i, j))

    # Only use pre-defined cpu pins (do not let board.csv create them).
    def find_pin_by_cpu_pin_name(self, cpu_pin_name, create=True):
        return super().find_pin_by_cpu_pin_name(cpu_pin_name, create=False)

    def cpu_table_size(self):
        return "{} * {}".format(NUM_PORTS, NUM_PINS_PER_PORT)


if __name__ == "__main__":
    AlifPinGenerator().main()



# test csr instructions

# CSR 0x340 is `mscratch`.  This test suite is only safe to run on a system
# where it is known that there is no other code that can read from or write
# to that register.  The qemu port is one such system, as the CSR is only
# accessed when a machine exception occurs, and at that point it doesn't matter
# anymore whether these tests are running or not.


@micropython.asm_rv32
def csr():
    li(a0, 0)
    csrrw(zero, zero, 0x340)  # All zeroes
    csrrs(a1, zero, 0x340)  # Read zeroes
    c_bnez(a1, end)
    addi(a0, a0, 1)
    li(a1, 0xA5A5A5A5)
    li(a2, 0x5A5A5A5A)
    csrrs(a2, a1, 0x340)  # Read zeroes, set 0xA5A5A5A5
    c_bnez(a2, end)
    addi(a0, a0, 1)
    csrrs(a3, zero, 0x340)  # Read 0xA5A5A5A5
    bne(a3, a1, end)
    addi(a0, a0, 1)
    li(a2, 0xF0F0F0F0)
    csrrc(zero, a2, 0x340)  # Clear upper half
    csrrs(a3, zero, 0x340)  # Read 0x05050505
    xori(a2, a2, -1)
    and_(a2, a1, a2)
    bne(a2, a3, end)
    addi(a0, a0, 1)
    label(end)


print(csr())


@micropython.asm_rv32
def csri():
    li(a0, 0)
    csrrwi(zero, 0x340, 15)  # Write 0xF
    csrrs(a1, zero, 0x340)  # Read 0xF
    csrrsi(a2, 0x340, 0)  # Read
    bne(a1, a2, end)
    addi(a0, a0, 1)
    csrrci(a2, 0x340, 0)  # Read
    bne(a1, a2, end)
    addi(a0, a0, 1)
    li(a2, 15)
    bne(a1, a2, end)
    addi(a0, a0, 1)
    csrrci(zero, 0x340, 1)  # Clear bit 1
    csrrs(a1, zero, 0x340)  # Read 0xE
    li(a2, 14)
    bne(a1, a2, end)
    addi(a0, a0, 1)
    csrrsi(zero, 0x340, 1)  # Set bit 1
    csrrs(a1, zero, 0x340)  # Read 0xF
    li(a2, 15)
    bne(a1, a2, end)
    addi(a0, a0, 1)
    label(end)


print(csri())



# test machine module

try:
    import machine

    machine.mem8
except:
    print("SKIP")
    raise SystemExit

import unittest


class Test(unittest.TestCase):
    def test_mem8_print(self):
        self.assertEqual(repr(machine.mem8), "<8-bit memory>")

    def test_alignment(self):
        with self.assertRaises(ValueError):
            machine.mem16[1]

        with self.assertRaises(ValueError):
            machine.mem16[1] = 1

    def test_operations(self):
        with self.assertRaises(TypeError):
            del machine.mem8[0]

        with self.assertRaises(TypeError):
            machine.mem8[0:1]

        with self.assertRaises(TypeError):
            machine.mem8[0:1] = 10

        with self.assertRaises(TypeError):
            machine.mem8["hello"]

        with self.assertRaises(TypeError):
            machine.mem8["hello"] = 10


if __name__ == "__main__":
    unittest.main()



# Test DTLS functionality including timeout handling

try:
    from tls import PROTOCOL_DTLS_CLIENT, PROTOCOL_DTLS_SERVER, SSLContext, CERT_NONE
    import io
except ImportError:
    print("SKIP")
    raise SystemExit


class DummySocket(io.IOBase):
    def __init__(self):
        self.write_buffer = bytearray()
        self.read_buffer = bytearray()

    def write(self, data):
        return len(data)

    def readinto(self, buf):
        # This is a placeholder socket that doesn't actually read anything
        # so the read buffer is always empty.
        return None

    def ioctl(self, req, arg):
        if req == 4:  # MP_STREAM_CLOSE
            return 0
        return -1


# Create dummy sockets for testing
server_socket = DummySocket()
client_socket = DummySocket()

# Wrap the DTLS Server
dtls_server_ctx = SSLContext(PROTOCOL_DTLS_SERVER)
dtls_server_ctx.verify_mode = CERT_NONE
dtls_server = dtls_server_ctx.wrap_socket(server_socket, do_handshake_on_connect=False)
print("Wrapped DTLS Server")

# Wrap the DTLS Client
dtls_client_ctx = SSLContext(PROTOCOL_DTLS_CLIENT)
dtls_client_ctx.verify_mode = CERT_NONE
dtls_client = dtls_client_ctx.wrap_socket(client_socket, do_handshake_on_connect=False)
print("Wrapped DTLS Client")

# Trigger the timing check multiple times with different elapsed times
for i in range(10):  # Try multiple iterations to hit the timing window
    dtls_client.write(b"test")
    data = dtls_server.read(1024)  # This should eventually hit the timing condition

print("OK")



# This is a regression test for https://github.com/micropython/micropython/issues/16619
# it runs in two parts by necessity:
#
# - This "part1" creates a non-terminating thread.
# - The test runner issues a soft reset, which will terminate that thread.
# - "part2" is the actual test, which is whether flash access works correctly
#   after the thread was terminated by soft reset.

import _thread


def infinite():
    while True:
        pass


_thread.start_new_thread(infinite, ())
print("Part 1 complete...")



# test branch instructions


@micropython.asm_rv32
def tbeq(a0):
    mv(a1, a0)

    li(a0, 10)
    li(a2, 1)
    beq(a1, a2, end)

    li(a0, 20)
    li(a2, 2)
    beq(a1, a2, end)

    li(a0, 30)
    li(a2, 3)
    beq(a1, a2, end)

    li(a0, 0)

    label(end)


print(tbeq(0))
print(tbeq(1))
print(tbeq(2))
print(tbeq(3))


@micropython.asm_rv32
def tbne(a0):
    mv(a1, a0)

    li(a0, 10)
    li(a2, 1)
    bne(a1, a2, end)

    li(a0, 20)
    li(a2, 2)
    bne(a1, a2, end)

    li(a0, 30)
    li(a2, 3)
    bne(a1, a2, end)

    li(a0, 0)

    label(end)


print(tbne(0))
print(tbne(1))
print(tbne(2))
print(tbne(3))


@micropython.asm_rv32
def tbgeu(a0):
    mv(a1, a0)

    li(a0, 1)
    li(a2, 2)
    bgeu(a1, a2, end)
    li(a0, 0)

    label(end)


print(tbgeu(0))
print(tbgeu(1))
print(tbgeu(2))
print(tbgeu(3))


@micropython.asm_rv32
def tbltu(a0):
    mv(a1, a0)

    li(a0, 1)
    li(a2, 2)
    bltu(a1, a2, end)
    li(a0, 0)

    label(end)


print(tbltu(0))
print(tbltu(1))
print(tbltu(2))
print(tbltu(3))


@micropython.asm_rv32
def tbge(a0):
    mv(a1, a0)

    li(a0, 1)
    li(a2, -2)
    bge(a1, a2, end)
    li(a0, 0)

    label(end)


print(tbge(-3))
print(tbge(-2))
print(tbge(-1))
print(tbge(0))


@micropython.asm_rv32
def tblt(a0):
    mv(a1, a0)

    li(a0, 1)
    li(a2, -2)
    blt(a1, a2, end)
    li(a0, 0)

    label(end)


print(tblt(-3))
print(tblt(-2))
print(tblt(-1))
print(tblt(0))


@micropython.asm_rv32
def tcbeqz(a0):
    mv(a1, a0)

    li(a0, 1)
    c_beqz(a1, end)
    li(a0, 0)

    label(end)


print(tcbeqz(0))
print(tcbeqz(1))
print(tcbeqz(2))
print(tcbeqz(3))


@micropython.asm_rv32
def tcbnez(a0):
    mv(a1, a0)

    li(a0, 1)
    c_bnez(a1, end)
    li(a0, 0)

    label(end)


print(tcbnez(0))
print(tcbnez(1))
print(tcbnez(2))
print(tcbnez(3))



# Basic Wi-Fi MAC layer test where one device creates an Access Point and the
# other device connects to it as a Station. Also tests channel assignment (where
# possible) and disconnection.

try:
    import network

    network.WLAN
except (ImportError, NameError):
    print("SKIP")
    raise SystemExit

import os
import sys
import time

CHANNEL = 8

# Note that on slower Wi-Fi stacks this bumps up against the run-multitests.py
# timeout which expects <10s between lines of output. We work around this by
# logging something half way through the wait_for loop...
CONNECT_TIMEOUT = 15000


def wait_for(test_func):
    has_printed = False
    start = time.ticks_ms()
    while not test_func():
        time.sleep(0.1)
        delta = time.ticks_diff(time.ticks_ms(), start)
        if not has_printed and delta > CONNECT_TIMEOUT / 2:
            print("...")
            has_printed = True
        elif delta > CONNECT_TIMEOUT:
            break

    if not has_printed:
        print("...")  # keep the output consistent

    return test_func()


# AP
def instance0():
    ap = network.WLAN(network.WLAN.IF_AP)
    ssid = "MP-test-" + os.urandom(6).hex()
    psk = "Secret-" + os.urandom(6).hex()

    # stop any previous activity
    network.WLAN(network.WLAN.IF_STA).active(False)
    ap.active(False)

    ap.active(True)
    ap.config(ssid=ssid, key=psk, channel=CHANNEL, security=network.WLAN.SEC_WPA_WPA2)

    # print("AP setup", ssid, psk)
    print("AP started")

    multitest.globals(SSID=ssid, PSK=psk)
    multitest.next()

    # Wait for station
    if not wait_for(ap.isconnected):
        raise RuntimeError("Timed out waiting for station, status ", ap.status())

    print("AP got station")
    time.sleep(
        3
    )  # depending on port, may still need to negotiate DHCP lease for STA to see connection

    print("AP disabling...")
    ap.active(False)


# STA
def instance1():
    sta = network.WLAN(network.WLAN.IF_STA)

    # stop any previous activity
    network.WLAN(network.WLAN.IF_AP).active(False)
    sta.active(False)

    multitest.next()
    ssid = SSID
    psk = PSK

    # print("STA setup", ssid, psk)

    sta.active(True)
    sta.connect(ssid, psk)

    print("STA connecting...")

    if not wait_for(sta.isconnected):
        raise RuntimeError("Timed out waiting to connect, status ", sta.status())

    print("STA connected")

    print("channel", sta.config("channel"))

    print("STA waiting for disconnect...")

    # Expect the AP to disconnect us immediately
    if not wait_for(lambda: not sta.isconnected()):
        raise RuntimeError("Timed out waiting for AP to disconnect us, status ", sta.status())

    print("STA disconnected")

    sta.active(False)



@micropython.asm_rv32
def lsl1(a0):
    slli(a0, a0, 1)


print(hex(lsl1(0x123)))


@micropython.asm_rv32
def lsl23(a0):
    slli(a0, a0, 23)


print(hex(lsl23(1)))


@micropython.asm_rv32
def lsr1(a0):
    srli(a0, a0, 1)


print(hex(lsr1(0x123)))


@micropython.asm_rv32
def lsr31(a0):
    srli(a0, a0, 31)


print(hex(lsr31(0x80000000)))


@micropython.asm_rv32
def asr1(a0):
    srai(a0, a0, 1)


print(hex(asr1(0x123)))


@micropython.asm_rv32
def asr31(a0):
    srai(a0, a0, 31)


print(hex(asr31(0x80000000)))


@micropython.asm_rv32
def clsl1(a0):
    c_slli(a0, 1)


print(hex(clsl1(0x123)))


@micropython.asm_rv32
def clsl23(a0):
    c_slli(a0, 23)


print(hex(clsl23(1)))


@micropython.asm_rv32
def clsr1(a0):
    c_srli(a0, 1)


print(hex(clsr1(0x123)))


@micropython.asm_rv32
def clsr31(a0):
    c_srli(a0, 31)


print(hex(clsr31(0x80000000)))


@micropython.asm_rv32
def casr1(a0):
    c_srai(a0, 1)


print(hex(casr1(0x123)))


@micropython.asm_rv32
def casr31(a0):
    c_srai(a0, 31)


print(hex(casr31(0x80000000)))


@micropython.asm_rv32
def lsl1r(a0):
    li(a1, 1)
    sll(a0, a0, a1)


print(hex(lsl1r(0x123)))


@micropython.asm_rv32
def lsr1r(a0):
    li(a1, 1)
    srl(a0, a0, a1)


print(hex(lsr1r(0x123)))


@micropython.asm_rv32
def asr1r(a0):
    li(a1, 1)
    sra(a0, a0, a1)


print(hex(asr1r(0x123)))



# test VFS functionality without any particular filesystem type

try:
    import os, vfs
except ImportError:
    print("SKIP")
    raise SystemExit
import errno


class Filesystem:
    def __init__(self, id, paths=[]):
        self.id = id
        self.paths = paths

    def mount(self, readonly, mksfs):
        print("mount", self)

    def umount(self):
        print("umount", self)

    def stat(self, path):
        if path in self.paths:
            return (0, 0, 0, 0, 0, 0, 0, 0, 0, 0)
        else:
            raise OSError

    statvfs = stat

    def open(self, path, mode):
        pass

    def __repr__(self):
        return "Filesystem(%d)" % self.id


# first we umount any existing mount points the target may have
try:
    vfs.umount("/")
except OSError:
    pass
for path in os.listdir("/"):
    vfs.umount("/" + path)


print(vfs.mount())

vfs.mount(Filesystem(1), "/foo")

print(vfs.mount())

vfs.mount(Filesystem(2), "/bar/baz")

print(vfs.mount())

vfs.mount(Filesystem(3), "/bar")

print(vfs.mount())

vfs.umount("/bar/baz")

print(vfs.mount())

vfs.mount(Filesystem(4), "/")

print(vfs.mount())



# test passing arguments


@micropython.asm_rv32
def arg0():
    c_li(a0, 1)


print(arg0())


@micropython.asm_rv32
def arg1(a0):
    addi(a0, a0, 1)


print(arg1(1))


@micropython.asm_rv32
def arg2(a0, a1):
    add(a0, a0, a1)


print(arg2(1, 2))


@micropython.asm_rv32
def arg3(a0, a1, a2):
    add(a0, a0, a1)
    add(a0, a0, a2)


print(arg3(1, 2, 3))


@micropython.asm_rv32
def arg4(a0, a1, a2, a3):
    add(a0, a0, a1)
    add(a0, a0, a2)
    add(a0, a0, a3)


print(arg4(1, 2, 3, 4))



# Test the marshal module, MicroPython-specific functionality.

try:
    import marshal
except ImportError:
    print("SKIP")
    raise SystemExit

import unittest


class Test(unittest.TestCase):
    def test_function_with_children(self):
        # Can't marshal a function with children (in this case the module has a child function f).
        code = compile("def f(): pass", "", "exec")
        with self.assertRaises(ValueError):
            marshal.dumps(code)


if __name__ == "__main__":
    unittest.main()



# Test machine.PWM, frequency and duty cycle (using machine.time_pulse_us).
#
# IMPORTANT: This test requires hardware connections: the PWM-output and pulse-input
# pins must be wired together (see the variable `pwm_pulse_pins`).

import sys
import time

try:
    from machine import time_pulse_us, Pin, PWM
except ImportError:
    print("SKIP")
    raise SystemExit

import unittest

pwm_freq_limit = 1000000
freq_margin_per_thousand = 0
duty_margin_per_thousand = 0
timing_margin_us = 5

# Configure pins based on the target.
if "esp32" in sys.platform:
    pwm_pulse_pins = ((4, 5),)
    freq_margin_per_thousand = 2
    duty_margin_per_thousand = 1
    timing_margin_us = 20
elif "esp8266" in sys.platform:
    pwm_pulse_pins = ((4, 5),)
    pwm_freq_limit = 1_000
    duty_margin_per_thousand = 3
    timing_margin_us = 50
elif "mimxrt" in sys.platform:
    if "Teensy" in sys.implementation._machine:
        # Teensy 4.x
        pwm_pulse_pins = (
            ("D0", "D1"),  # FLEXPWM X and UART 1
            ("D2", "D3"),  # FLEXPWM A/B
            ("D11", "D12"),  # QTMR and MOSI/MISO of SPI 0
        )
    else:
        pwm_pulse_pins = (("D0", "D1"),)
elif "rp2" in sys.platform:
    pwm_pulse_pins = (("GPIO0", "GPIO1"),)
elif "samd" in sys.platform:
    pwm_pulse_pins = (("D0", "D1"),)
    if "SAMD21" in sys.implementation._machine:
        # MCU is too slow to capture short pulses.
        pwm_freq_limit = 2_000
else:
    print("Please add support for this test on this platform.")
    raise SystemExit


# Test a specific frequency and duty cycle.
def _test_freq_duty(self, pulse_in, pwm, freq, duty_u16):
    print("freq={:<5} duty_u16={:<5} :".format(freq, duty_u16), end="")

    # Check configured freq/duty_u16 is within error bound.
    freq_error = abs(pwm.freq() - freq) * 1000 // freq
    duty_error = abs(pwm.duty_u16() - duty_u16) * 1000 // (duty_u16 or 1)
    print(" freq={} freq_er={}".format(pwm.freq(), freq_error), end="")
    print(" duty={} duty_er={}".format(pwm.duty_u16(), duty_error), end="")
    print(" :", end="")
    self.assertLessEqual(freq_error, freq_margin_per_thousand)
    self.assertLessEqual(duty_error, duty_margin_per_thousand)

    # Calculate expected timing.
    expected_total_us = 1_000_000 // freq
    expected_high_us = expected_total_us * duty_u16 // 65535
    expected_low_us = expected_total_us - expected_high_us
    expected_us = (expected_low_us, expected_high_us)
    timeout = 2 * expected_total_us

    # Wait for output to settle.
    time_pulse_us(pulse_in, 0, timeout)
    time_pulse_us(pulse_in, 1, timeout)

    if duty_u16 == 0 or duty_u16 == 65535:
        # Expect a constant output level.
        no_pulse = (
            time_pulse_us(pulse_in, 0, timeout) < 0 and time_pulse_us(pulse_in, 1, timeout) < 0
        )
        self.assertTrue(no_pulse)
        if expected_high_us == 0:
            # Expect a constant low level.
            self.assertEqual(pulse_in(), 0)
        else:
            # Expect a constant high level.
            self.assertEqual(pulse_in(), 1)
    else:
        # Test timing of low and high pulse.
        n_averaging = 10
        for level in (0, 1):
            t = 0
            time_pulse_us(pulse_in, level, timeout)
            for _ in range(n_averaging):
                t += time_pulse_us(pulse_in, level, timeout)
            t //= n_averaging
            expected = expected_us[level]
            print(" level={} timing_er={}".format(level, abs(t - expected)), end="")
            self.assertLessEqual(abs(t - expected), timing_margin_us)

    print()


# Test a specific frequency with multiple duty cycles.
def _test_freq(self, freq):
    print()
    self.pwm.freq(freq)
    for duty in (0, 10, 25, 50, 75, 90, 100):
        duty_u16 = duty * 65535 // 100
        if sys.platform == "esp32":
            # TODO why is this bit needed to get it working on esp32?
            self.pwm.init(freq=freq, duty_u16=duty_u16)
            time.sleep(0.1)
        self.pwm.duty_u16(duty_u16)
        _test_freq_duty(self, self.pulse_in, self.pwm, freq, duty_u16)


# Given a set of pins, this test class will test multiple frequencies and duty cycles.
class TestBase:
    @classmethod
    def setUpClass(cls):
        print("set up pins:", cls.pwm_pin, cls.pulse_pin)
        cls.pwm = PWM(cls.pwm_pin)
        cls.pulse_in = Pin(cls.pulse_pin, Pin.IN)

    @classmethod
    def tearDownClass(cls):
        cls.pwm.deinit()

    def test_freq_50(self):
        _test_freq(self, 50)

    def test_freq_100(self):
        _test_freq(self, 100)

    def test_freq_500(self):
        _test_freq(self, 500)

    def test_freq_1000(self):
        _test_freq(self, 1000)

    @unittest.skipIf(pwm_freq_limit < 2000, "frequency too high")
    def test_freq_2000(self):
        _test_freq(self, 2000)

    @unittest.skipIf(pwm_freq_limit < 5000, "frequency too high")
    def test_freq_5000(self):
        _test_freq(self, 5000)

    @unittest.skipIf(pwm_freq_limit < 10000, "frequency too high")
    def test_freq_10000(self):
        _test_freq(self, 10000)


# Generate test classes, one for each set of pins to test.
for pwm, pulse in pwm_pulse_pins:
    cls_name = "Test_{}_{}".format(pwm, pulse)
    globals()[cls_name] = type(
        cls_name, (TestBase, unittest.TestCase), {"pwm_pin": pwm, "pulse_pin": pulse}
    )

if __name__ == "__main__":
    unittest.main()



@micropython.asm_rv32
def asm_sum_words(a0, a1):
    # a0 = len
    # a1 = ptr
    # a2 = sum
    # a3 = dummy
    li(a2, 0)

    jal(zero, loop_entry)

    label(loop1)
    lw(a3, 0(a1))
    add(a2, a2, a3)

    addi(a1, a1, 4)
    addi(a0, a0, -1)

    label(loop_entry)
    bne(a0, zero, loop1)

    mv(a0, a2)


@micropython.asm_rv32
def asm_sum_bytes(a0, a1):
    # a0 = len
    # a1 = ptr
    # a2 = sum
    # a3 = dummy
    li(a2, 0)

    jal(zero, loop_entry)

    label(loop1)
    lbu(a3, 0(a1))
    add(a2, a2, a3)

    addi(a1, a1, 1)
    addi(a0, a0, -1)

    label(loop_entry)
    bne(a0, zero, loop1)

    mv(a0, a2)


import array

b = array.array("l", (100, 200, 300, 400))
n = asm_sum_words(len(b), b)
print(b, n)

b = array.array("b", (10, 20, 30, 40, 50, 60, 70, 80))
n = asm_sum_bytes(len(b), b)
print(b, n)

b = b"\x01\x02\x03\x04"
n = asm_sum_bytes(len(b), b)
print(b, n)



@micropython.asm_rv32
def sdiv(a0, a1):
    div(a0, a0, a1)


@micropython.asm_rv32
def udiv(a0, a1):
    divu(a0, a0, a1)


@micropython.asm_rv32
def srem(a0, a1):
    rem(a0, a0, a1)


@micropython.asm_rv32
def urem(a0, a1):
    remu(a0, a0, a1)


print(sdiv(1234, 3))
print(sdiv(-1234, 3))
print(sdiv(1234, -3))
print(sdiv(-1234, -3))

print(udiv(1234, 3))
print(udiv(0xFFFFFFFF, 0x7FFFFFFF))
print(udiv(0xFFFFFFFF, 0xFFFFFFFF))

print(srem(1234, 3))
print(srem(-1234, 3))
print(srem(1234, -3))
print(srem(-1234, -3))

print(urem(1234, 3))
print(urem(0xFFFFFFFF, 0x7FFFFFFF))
print(urem(0xFFFFFFFF, 0xFFFFFFFF))


@micropython.asm_rv32
def m1(a0, a1):
    mul(a0, a0, a1)


@micropython.asm_rv32
def m2(a0, a1):
    mulh(a0, a0, a1)


@micropython.asm_rv32
def m3(a0, a1):
    mulhu(a0, a0, a1)


@micropython.asm_rv32
def m4(a0, a1):
    mulhsu(a0, a0, a1)


print(m1(0xFFFFFFFF, 2))
print(m2(0xFFFFFFFF, 0xFFFFFFF0))
print(m3(0xFFFFFFFF, 0xFFFFFFF0))
print(m4(0xFFFFFFFF, 0xFFFFFFF0))



from pyb import CAN
import time
import errno

# Test for the filtering capabilities for RX FIFO 0 and 1.


# Receiver
def instance0():
    # Configure to receive standard frames (in a range) on FIFO 0
    # and extended frames (in a range) on FIFO 1.
    if hasattr(CAN, "MASK"):
        # FD-CAN has independent filter banks for standard and extended IDs
        can.setfilter(0, CAN.MASK, 0, (0x300, 0x700), extframe=False)
        can.setfilter(0, CAN.MASK, 1, (0x3000, 0x7000), extframe=True)
    else:
        # pyb.CAN only supports MASK32 for extended ids
        can.setfilter(0, CAN.MASK16, 0, (0x300, 0x700, 0x300, 0x700), extframe=False)
        can.setfilter(1, CAN.MASK32, 1, (0x3000, 0x7000), extframe=True)

    multitest.next()
    multitest.wait("sender ready")
    multitest.broadcast("receiver ready")
    for i in range(3):
        print(i)
        print("fifo0", can.recv(0, timeout=200))
        print("fifo1", can.recv(1, timeout=200))

    try:
        can.recv(0, timeout=100)  # should time out
    except OSError as e:
        assert e.errno == errno.ETIMEDOUT
        print("Timed out as expected")


# Sender
def instance1():
    multitest.next()
    multitest.broadcast("sender ready")
    multitest.wait("receiver ready")

    for i in range(3):
        print(i)
        can.send(bytes([i, 3] * i), 0x345)
        can.send(bytes([0xEE] * i), 0x3700 + i, extframe=True)
        can.send(b"abcdef", 0x123)  # matches no filter, expect ACKed but not received
        time.sleep_ms(5)  # avoid flooding either our or the receiver's FIFO


can = CAN(1, CAN.NORMAL, baudrate=500_000, sample_point=75)



TEMPLATE3 = """
@micropython.asm_rv32
def f():
    {}({}, {}, {})
"""

TEMPLATE2 = """
@micropython.asm_rv32
def f():
    {}({}, {})
"""

TEMPLATE1 = """
@micropython.asm_rv32
def f():
    {}({})
"""


REGISTERS = [
    "zero",
    "s0",
    "s1",
    "s2",
    "s3",
    "s4",
    "s5",
    "s6",
    "s7",
    "s8",
    "s9",
    "s10",
    "s11",
    "a0",
    "a1",
    "a2",
    "a3",
    "a4",
    "a5",
    "a6",
    "a7",
    "tp",
    "gp",
    "sp",
    "ra",
    "t0",
    "t1",
    "t2",
    "t3",
    "t4",
    "t5",
    "t6",
    "x0",
    "x1",
    "x2",
    "x3",
    "x4",
    "x5",
    "x6",
    "x7",
    "x8",
    "x9",
    "x10",
    "x11",
    "x12",
    "x13",
    "x14",
    "x15",
    "x16",
    "x17",
    "x18",
    "x19",
    "x20",
    "x21",
    "x22",
    "x23",
    "x24",
    "x25",
    "x26",
    "x27",
    "x28",
    "x29",
    "x30",
    "x31",
]


def harness(opcode, fragment, tag):
    try:
        exec(fragment)
    except SyntaxError:
        print(tag, opcode)


for opcode in ("slli", "srli", "srai"):
    harness(opcode, TEMPLATE3.format(opcode, "a0", "a0", -1), "-")
    harness(opcode, TEMPLATE3.format(opcode, "a0", "a0", 33), "+")

for opcode in ("c_slli", "c_srli", "c_srai"):
    harness(opcode, TEMPLATE2.format(opcode, "a0", -1), "-")
    harness(opcode, TEMPLATE2.format(opcode, "a0", 33), "+")

harness("c_slli", TEMPLATE2.format("c_slli", "zero", 0), "0")
harness("c_slli", TEMPLATE2.format("c_slli", "x0", 0), "0")

for opcode in ("c_srli", "c_srai"):
    for register in REGISTERS:
        harness(opcode, TEMPLATE2.format(opcode, register, 0), register)

for opcode in ("c_mv", "c_add"):
    harness(opcode, TEMPLATE2.format(opcode, "a0", "zero"), "0l")
    harness(opcode, TEMPLATE2.format(opcode, "zero", "a0"), "0r")
    harness(opcode, TEMPLATE2.format(opcode, "zero", "zero"), "0b")

harness("c_jr", TEMPLATE1.format("c_jr", "zero"), "0")

for opcode in ("addi", "andi", "ori", "slti", "sltiu", "xori"):
    harness(opcode, TEMPLATE3.format(opcode, "a0", "a0", 0x7FF), ">=s")
    harness(opcode, TEMPLATE3.format(opcode, "a0", "a0", 0x800), ">s")
    harness(opcode, TEMPLATE3.format(opcode, "a0", "a0", -2048), "<=s")
    harness(opcode, TEMPLATE3.format(opcode, "a0", "a0", -2049), "<s")

for opcode in ("lb", "lbu", "lh", "lhu", "lw", "sb", "sh", "sw"):
    TEMPLATE = """
@micropython.asm_rv32
def f():
    {}(a0, {}(a0))
"""
    harness(opcode, TEMPLATE.format(opcode, 0x7FF), ">=s")
    harness(opcode, TEMPLATE.format(opcode, 0x800), ">s")
    harness(opcode, TEMPLATE.format(opcode, -2048), "<=s")
    harness(opcode, TEMPLATE.format(opcode, -2049), "<s")

harness("c_addi", TEMPLATE2.format("c_andi", "zero", 0), "00")
harness("c_addi", TEMPLATE2.format("c_andi", "zero", 512), ">0")
harness("c_addi", TEMPLATE2.format("c_andi", "zero", -512), "<0")
harness("c_addi", TEMPLATE2.format("c_andi", "s0", 0), "s0")
harness("c_addi", TEMPLATE2.format("c_andi", "s0", -100), "<s")
harness("c_addi", TEMPLATE2.format("c_andi", "s0", 100), ">s")

harness("c_andi", TEMPLATE2.format("c_andi", "zero", 0), "00")
harness("c_andi", TEMPLATE2.format("c_andi", "zero", 512), ">0")
harness("c_andi", TEMPLATE2.format("c_andi", "zero", -512), "<0")
harness("c_andi", TEMPLATE2.format("c_andi", "s0", 0), "s0")
harness("c_andi", TEMPLATE2.format("c_andi", "s0", -100), "<s")
harness("c_andi", TEMPLATE2.format("c_andi", "s0", 100), ">s")

C_REGISTERS = (
    "a0",
    "a1",
    "a2",
    "a3",
    "a4",
    "a5",
    "s0",
    "s1",
    "x8",
    "x9",
    "x10",
    "x11",
    "x12",
    "x13",
    "x14",
    "x15",
)

for opcode in ("c_and", "c_or", "c_xor"):
    for source in REGISTERS:
        for destination in REGISTERS:
            if source in C_REGISTERS and destination in C_REGISTERS:
                try:
                    exec(
                        """
@micropython.asm_rv32
def f():
    {}({}, {})
""".format(opcode, source, destination)
                    )
                except SyntaxError:
                    print(source, destination, opcode)
            else:
                try:
                    exec(
                        """
@micropython.asm_rv32
def f():
    {}({}, {})
""".format(opcode, source, destination)
                    )
                    print(source, destination, opcode)
                except SyntaxError:
                    pass
    print(opcode)

for opcode in ("c_lw", "c_sw"):
    TEMPLATE = """
@micropython.asm_rv32
def f():
    {}(a0, {}(a0))
"""
    harness(opcode, TEMPLATE.format(opcode, 60), ">=s")
    harness(opcode, TEMPLATE.format(opcode, 61), ">s")
    harness(opcode, TEMPLATE.format(opcode, -60), "<=s")
    harness(opcode, TEMPLATE.format(opcode, -61), "<s")



# Test deflate.DeflateIO compression, with out-of-memory errors.

try:
    # Check if deflate is available.
    import deflate
    import io
except ImportError:
    print("SKIP")
    raise SystemExit

# Check if compression is enabled.
if not hasattr(deflate.DeflateIO, "write"):
    print("SKIP")
    raise SystemExit

# Create a compressor object.
b = io.BytesIO()
g = deflate.DeflateIO(b, deflate.RAW, 15)

# Then, use up most of the heap.
l = []
while True:
    try:
        l.append(bytearray(1000))
    except:
        break
l.pop()

# Try to compress.  This will try to allocate a large window and fail.
try:
    g.write('test')
except MemoryError:
    print("MemoryError")

# Should still be able to close the stream.
g.close()

# The underlying output stream should be unchanged.
print(b.getvalue())



# Ensure that SSLSockets can be allocated sequentially
# without running out of available memory.
try:
    import io
    import tls
except ImportError:
    print("SKIP")
    raise SystemExit

import unittest


class TestSocket(io.IOBase):
    def write(self, buf):
        return len(buf)

    def readinto(self, buf):
        return 0

    def ioctl(self, cmd, arg):
        return 0

    def setblocking(self, value):
        pass


ITERS = 128


class TLSNoLeaks(unittest.TestCase):
    def test_unique_context(self):
        for n in range(ITERS):
            print(n)
            s = TestSocket()
            ctx = tls.SSLContext(tls.PROTOCOL_TLS_CLIENT)
            ctx.verify_mode = tls.CERT_NONE
            s = ctx.wrap_socket(s, do_handshake_on_connect=False)

    def test_shared_context(self):
        # Single SSLContext, multiple sockets
        ctx = tls.SSLContext(tls.PROTOCOL_TLS_CLIENT)
        ctx.verify_mode = tls.CERT_NONE
        for n in range(ITERS):
            print(n)
            s = TestSocket()
            s = ctx.wrap_socket(s, do_handshake_on_connect=False)


if __name__ == "__main__":
    unittest.main()



# Test DTLS server and client, sending a small amount of data between them.

try:
    import socket
    import tls
except ImportError:
    print("SKIP")
    raise SystemExit

PORT = 8000

# These are test certificates. See tests/README.md for details.
certfile = "ec_cert.der"
keyfile = "ec_key.der"

try:
    with open(certfile, "rb") as cf:
        cert = cadata = cf.read()
    with open(keyfile, "rb") as kf:
        key = kf.read()
except OSError:
    print("SKIP")
    raise SystemExit


# DTLS server.
def instance0():
    multitest.globals(IP=multitest.get_network_ip())

    # Create a UDP socket and bind it to accept incoming connections.
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    s.bind(socket.getaddrinfo("0.0.0.0", PORT)[0][-1])

    multitest.next()

    # Wait for the client to connect.
    data, client_addr = s.recvfrom(1)
    print("incoming connection", data)

    # Connect back to the client, so the UDP socket can be used like a stream.
    s.connect(client_addr)

    # Create the DTLS context and load the certificate.
    ctx = tls.SSLContext(tls.PROTOCOL_DTLS_SERVER)
    ctx.load_cert_chain(cert, key)

    # Wrap the UDP socket in server mode.
    print("wrap socket")
    s = ctx.wrap_socket(s, server_side=1)

    # Transfer some data.
    for _ in range(4):
        print(s.recv(16))
        s.send(b"server to client")

    # Close the DTLS and UDP connection.
    s.close()


# DTLS client.
def instance1():
    multitest.next()

    # Create a UDP socket and connect to the server.
    addr = socket.getaddrinfo(IP, PORT)[0][-1]
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    print("connect")
    s.connect(addr)

    # Send one byte to indicate a connection, and so the server can obtain our address.
    s.write("X")

    # Create a DTLS context and load the certificate.
    ctx = tls.SSLContext(tls.PROTOCOL_DTLS_CLIENT)
    ctx.verify_mode = tls.CERT_REQUIRED
    ctx.load_verify_locations(cadata)

    # Wrap the UDP socket.
    print("wrap socket")
    s = ctx.wrap_socket(s, server_hostname="micropython.local")

    # Transfer some data.
    for _ in range(4):
        s.send(b"client to server")
        print(s.recv(16))

    # Close the DTLS and UDP connection.
    s.close()



# test arithmetic opcodes


@micropython.asm_rv32
def f1():
    li(a0, 0x100)
    li(a1, 1)
    add(a0, a0, a1)
    addi(a0, a0, 1)
    addi(a0, a0, -2)
    sub(a0, a0, a1)
    c_add(a0, a1)
    c_addi(a0, -1)
    c_sub(a0, a1)


print(hex(f1()))


@micropython.asm_rv32
def f2():
    li(a0, 0x10FF)
    li(a1, 1)
    and_(a2, a0, a1)
    andi(a3, a0, 0x10)
    or_(a2, a2, a3)
    ori(a2, a2, 8)
    li(a1, 0x200)
    c_or(a2, a1)
    li(a1, 0xF0)
    mv(a0, a2)
    c_and(a0, a1)
    li(a1, 0x101)
    xor(a0, a0, a1)
    xori(a0, a0, 0x101)
    c_xor(a0, a1)


print(hex(f2()))


@micropython.asm_rv32
def f3(a0, a1):
    slt(a0, a0, a1)


print(f3(0xFFFFFFF0, 0xFFFFFFF1))
print(f3(0x0, 0xFFFFFFF1))
print(f3(0xFFFFFFF1, 0xFFFFFFF1))
print(f3(0xFFFFFFF1, 0xFFFFFFF0))


@micropython.asm_rv32
def f4(a0, a1):
    sltu(a0, a0, a1)


print(f3(0xFFFFFFF0, 0xFFFFFFF1))
print(f3(0x0, 0xFFFFFFF1))
print(f3(0xFFFFFFF1, 0xFFFFFFF1))
print(f3(0xFFFFFFF1, 0xFFFFFFF0))


@micropython.asm_rv32
def f5(a0):
    slti(a0, a0, -2)


print(f5(-1))
print(f5(-3))


@micropython.asm_rv32
def f6(a0):
    sltiu(a0, a0, -2)


print(f6(-1))
print(f6(-3))



# test stack manipulation


@micropython.asm_rv32
def l():
    li(a2, 3)
    addi(sp, sp, -8)
    li(a0, 0x123)
    sw(a0, 0(sp))
    addi(a1, a0, 0x111)
    sw(a1, 4(sp))
    c_lwsp(a3, 0)
    bne(a3, a0, END)
    addi(a2, a2, -1)
    c_lwsp(a3, 4)
    bne(a3, a1, END)
    addi(a2, a2, -1)
    label(END)
    addi(sp, sp, 8)
    mv(a0, a2)


print(l())


@micropython.asm_rv32
def s():
    li(a2, 3)
    addi(sp, sp, -8)
    li(a0, 0x123)
    c_swsp(a0, 0)
    addi(a1, a0, 0x111)
    c_swsp(a1, 4)
    lw(a3, 0(sp))
    bne(a3, a0, END)
    addi(a2, a2, -1)
    lw(a3, 4(sp))
    bne(a3, a1, END)
    addi(a2, a2, -1)
    label(END)
    addi(sp, sp, 8)
    mv(a0, a2)


print(s())


@micropython.asm_rv32
def a():
    li(a2, 3)
    addi(sp, sp, -8)
    li(a0, 0x123)
    sw(a0, 0(sp))
    addi(a1, a0, 0x111)
    sw(a1, 4(sp))
    c_addi4spn(a3, 4)
    lw(a3, -4(a3))
    bne(a3, a0, END)
    addi(a2, a2, -1)
    label(END)
    addi(sp, sp, 8)
    mv(a0, a2)


print(a())



# Test Loop.new_event_loop()

try:
    import asyncio
except ImportError:
    print("SKIP")
    raise SystemExit

# CPython 3.12 deprecated calling get_event_loop() when there is no current event
# loop, so to make this test run on CPython requires setting the event loop.
if hasattr(asyncio, "set_event_loop"):
    asyncio.set_event_loop(asyncio.new_event_loop())


def exception_handler(loop, context):
    # This is a workaround for a difference between CPython and MicroPython: if
    # a CPython event loop is closed while there are tasks pending (i.e. not finished)
    # on it, then the task will log an error. MicroPython does not log this error.
    if context.get("message", "") == "Task was destroyed but it is pending!":
        pass
    else:
        loop.default_exception_handler(context)


async def task():
    for i in range(4):
        print("task", i)
        await asyncio.sleep(0)
        await asyncio.sleep(0)


async def main():
    print("start")
    loop.create_task(task())
    await asyncio.sleep(0)  # yields, meaning new task will run once
    print("stop")
    loop.stop()


# Use default event loop to run some tasks
loop = asyncio.get_event_loop()
loop.set_exception_handler(exception_handler)
loop.create_task(main())
loop.run_forever()
loop.close()

# Create new event loop, old one should not keep running
loop = asyncio.new_event_loop()
loop.set_exception_handler(exception_handler)
loop.create_task(main())
loop.run_forever()
loop.close()



# Test MicroPython-specific restrictions of function.__code__ attribute.

try:
    (lambda: 0).__code__
except AttributeError:
    print("SKIP")
    raise SystemExit

import micropython


@micropython.native
def f_native():
    pass


# Can't access __code__ when function is native code.
try:
    f_native.__code__
except AttributeError:
    print("AttributeError")



@micropython.asm_rv32
def f1():
    li(a0, 0)
    la(a1, END)
    c_jr(a1)
    c_addi(a0, 1)
    c_addi(a0, 1)
    c_addi(a0, 1)
    c_addi(a0, 1)
    c_addi(a0, 1)
    label(END)


print(f1())


@micropython.asm_rv32
def f2():
    addi(sp, sp, -4)
    c_swsp(ra, 0)
    li(ra, 0)
    li(a0, 0)
    c_jal(END)
    c_addi(a0, 1)
    c_addi(a0, 1)
    c_addi(a0, 1)
    c_addi(a0, 1)
    c_addi(a0, 1)
    label(END)
    bne(ra, zero, SUCCESS)
    c_addi(a0, 2)
    label(SUCCESS)
    c_lwsp(ra, 0)
    addi(sp, sp, 4)


print(f2())


@micropython.asm_rv32
def f3():
    li(a0, 0)
    c_j(END)
    c_addi(a0, 1)
    c_addi(a0, 1)
    c_addi(a0, 1)
    c_addi(a0, 1)
    c_addi(a0, 1)
    label(END)


print(f3())


@micropython.asm_rv32
def f4():
    addi(sp, sp, -4)
    c_swsp(ra, 0)
    li(ra, 0)
    li(a0, 0)
    la(a1, END)
    c_jalr(a1)
    c_addi(a0, 1)
    c_addi(a0, 1)
    c_addi(a0, 1)
    c_addi(a0, 1)
    c_addi(a0, 1)
    label(END)
    bne(ra, zero, SUCCESS)
    c_addi(a0, 2)
    label(SUCCESS)
    c_lwsp(ra, 0)
    addi(sp, sp, 4)


print(f4())


@micropython.asm_rv32
def f5():
    li(a0, 0)
    li(a1, 0)
    jal(a1, END)
    c_addi(a0, 1)
    c_addi(a0, 1)
    c_addi(a0, 1)
    c_addi(a0, 1)
    c_addi(a0, 1)
    label(END)
    bne(a1, zero, SUCCESS)
    c_addi(a0, 2)
    label(SUCCESS)


print(f5())


@micropython.asm_rv32
def f6():
    li(a0, 0)
    la(a1, JUMP)
    li(a2, 0)
    jalr(a2, a1, 10)
    label(JUMP)
    c_addi(a0, 1)
    c_addi(a0, 1)
    c_addi(a0, 1)
    c_addi(a0, 1)
    c_addi(a0, 1)
    bne(a2, zero, SUCCESS)
    c_addi(a0, 2)
    label(SUCCESS)


print(f6())



# Test that ESP-NOW picks up the channel configuration for STA
# mode on ESP32.
#
# Note that setting the channel on a peer in ESP-NOW on modern ESP-IDF only
# checks it against the configured channel, it doesn't ever change the radio
# channel
import sys
import time

try:
    import network
    import espnow
except ImportError:
    print("SKIP")
    raise SystemExit

# ESP8266 doesn't support config('channel') on the STA interface,
# and the channel parameter to add_peer doesn't appear to set the
# channel either.
if sys.platform == "esp8266":
    print("SKIP")
    raise SystemExit


timeout_ms = 1000
default_pmk = b"MicroPyth0nRules"

CHANNEL = 3
WRONG_CHANNEL = 8


def init_sta():
    sta = network.WLAN(network.WLAN.IF_STA)
    e = espnow.ESPNow()
    e.active(True)
    sta.active(True)
    sta.disconnect()  # Force AP disconnect for any saved config, important so the channel doesn't change
    sta.config(channel=CHANNEL)
    e.set_pmk(default_pmk)
    return sta, e


# Receiver
def instance0():
    sta, e = init_sta()
    multitest.globals(PEER=sta.config("mac"))
    multitest.next()
    print(sta.config("channel"))
    while True:
        peer, msg = e.recv(timeout_ms)
        if peer is None:
            print("Timeout")
            break
        print(msg)
    e.active(False)


# Sender
def instance1():
    sta, e = init_sta()
    multitest.next()
    peer = PEER

    # both instances set channel via sta.config(), above
    msg = b"sent to right channel 1"
    e.add_peer(peer, channel=CHANNEL)
    for _ in range(3):
        e.send(peer, msg)
    e.del_peer(peer)
    print(sta.config("channel"))

    sta.config(channel=WRONG_CHANNEL)
    msg = b"sent to wrong channel"
    e.add_peer(peer, channel=WRONG_CHANNEL)
    for _ in range(3):
        e.send(peer, msg)
    e.del_peer(peer)
    print(sta.config("channel"))

    # switching back to the correct channel should also work
    sta.config(channel=CHANNEL)
    msg = b"sent to right channel 2"
    e.add_peer(peer, channel=CHANNEL)
    for _ in range(3):
        e.send(peer, msg)
    e.del_peer(peer)
    print(sta.config("channel"))

    e.active(False)



import sys, os, alif


bdev = alif.Flash()
if hasattr(alif, "usb_msc"):
    try:
        # This may fail on VfsFat construction, or mount.
        os.mount(os.VfsFat(bdev), "/flash")
    except:
        os.VfsFat.mkfs(bdev)
        os.mount(os.VfsFat(bdev), "/flash")
else:
    try:
        os.mount(os.VfsLfs2(bdev, progsize=256), "/flash")
    except:
        os.VfsLfs2.mkfs(bdev, progsize=256)
        os.mount(os.VfsLfs2(bdev, progsize=256), "/flash")

sys.path.append("/flash")
sys.path.append("/flash/lib")
os.chdir("/flash")

del sys, os, alif, bdev



# This is part2 of a two-part regression test, see part1
# for details of what's expected.
import os

FILENAME = "/rp2_thread_reset_test.txt"

print("Starting")
with open(FILENAME, "w") as f:
    f.write("test")
print("Written")
os.unlink(FILENAME)
print("Removed")



# Test MicroPython-specific restrictions of function.__code__ attribute.

try:
    (lambda: 0).__code__
except AttributeError:
    print("SKIP")
    raise SystemExit


def f_with_children():
    def g():
        pass


# Can't access __code__ when function has children.
try:
    f_with_children.__code__
except AttributeError:
    print("AttributeError")



# Test the marshal module, basic functionality.

try:
    import marshal

    (lambda: 0).__code__
except (AttributeError, ImportError):
    print("SKIP")
    raise SystemExit

ftype = type(lambda: 0)

# Test basic dumps and loads.
print(ftype(marshal.loads(marshal.dumps((lambda: a).__code__)), {"a": 4})())

# Test dumps of a result from compile().
ftype(marshal.loads(marshal.dumps(compile("print(a)", "", "exec"))), {"print": print, "a": 5})()

# Test marshalling a function with arguments.
print(ftype(marshal.loads(marshal.dumps((lambda x, y: x + y).__code__)), {})(1, 2))

# Test marshalling a function with default arguments.
print(ftype(marshal.loads(marshal.dumps((lambda x=0: x).__code__)), {})("arg"))

# Test marshalling a function containing constant objects (a tuple).
print(ftype(marshal.loads(marshal.dumps((lambda: (None, ...)).__code__)), {})())

# Test instantiating multiple code's with different globals dicts.
code = marshal.loads(marshal.dumps((lambda: a).__code__))
f1 = ftype(code, {"a": 1})
f2 = ftype(code, {"a": 2})
print(f1(), f2())

# Test unmarshallable object.
try:
    marshal.dumps(type)
except ValueError:
    print("ValueError")



# MIT license; Copyright (c) 2024 OpenMV LLC.
#
# Note: this test completely erases the filesystem!
# It is intended to be run manually.

import alif
import hashlib
import os
import sys
import time
import vfs

hash_algo = "sha256"


def flash_make_filesystem():
    try:
        vfs.umount("/flash")
    except:
        pass
    bdev = alif.Flash()
    vfs.VfsFat.mkfs(bdev)
    vfs.mount(vfs.VfsFat(bdev), "/flash")
    sys.path.append("/flash")
    sys.path.append("/flash/lib")
    os.chdir("/flash")


def flash_block_test():
    try:
        vfs.umount("/flash")
    except:
        pass
    dev = alif.Flash()

    data512 = os.urandom(512)
    buf512 = bytearray(512)
    block_numbers = tuple(range(32)) + tuple(range(250, 266))

    print("Block read/write integrity: ", end="")
    ok = True
    for block_n in block_numbers:
        dev.writeblocks(block_n, data512)
        dev.readblocks(block_n, buf512)
        if buf512 != data512:
            ok = False
    print(ok)

    print("Block read back integrity: ", end="")
    ok = True
    for block_n in block_numbers:
        dev.readblocks(block_n, buf512)
        if buf512 != data512:
            ok = False
    print(ok)

    N = 16 * 1024
    data_big = os.urandom(N)
    t0 = time.ticks_us()
    dev.writeblocks(0, data_big)
    dt = time.ticks_diff(time.ticks_us(), t0)
    print(f"Block write speed: {len(data_big) / 1024 / dt * 1_000_000} KiB/sec")

    buf_big = bytearray(N)
    t0 = time.ticks_us()
    dev.readblocks(0, buf_big)
    dt = time.ticks_diff(time.ticks_us(), t0)
    print(f"Block read speed:  {len(buf_big) / 1024 / dt * 1_000_000} KiB/sec")

    if buf_big != data_big:
        raise RuntimeError("big block read-back failed")

    try:
        import uctypes

        xip = memoryview(dev)
    except:
        xip = None
    if xip is not None:
        t0 = time.ticks_us()
        buf_big[:] = xip[: len(buf_big)]
        dt = time.ticks_diff(time.ticks_us(), t0)
        print(f"XIP read speed:    {len(buf_big) / 1024 / dt * 1_000_000} KiB/sec")

        if buf_big != data_big:
            raise RuntimeError("XIP read-back failed")
        for i in range(len(buf_big)):
            if xip[i] != data_big[i]:
                raise RuntimeError("XIP byte-wise read-back failed")


def flash_write_verify(path, size=1024 * 1024, block_size=1024):
    hash_sum = getattr(hashlib, hash_algo)()
    block_count = size // block_size

    print("-" * 16)
    print(f"Writing file {size=} {block_size=}")
    t0 = time.ticks_ms()
    total_size = 0
    with open(path, "wb") as file:
        for i in range(block_count):
            buf = os.urandom(block_size)
            # Update digest
            hash_sum.update(buf)
            total_size += file.write(buf)
            if i % (block_count // 16) == 0:
                print(f"{i}/{block_count}", end="\r")
    dt = time.ticks_diff(time.ticks_ms(), t0)
    print(f"Flash write finished:       {total_size / 1024 / dt * 1000} KiB/sec")
    digest = hash_sum.digest()

    print("Reading file... ", end="")
    hash_sum = getattr(hashlib, hash_algo)()
    buf = bytearray(block_size)
    t0 = time.ticks_ms()
    total_size = 0
    with open(path, "rb") as file:
        for i in range(block_count):
            total_size += file.readinto(buf)
    dt = time.ticks_diff(time.ticks_ms(), t0)
    print(f"finished:   {total_size / 1024 / dt * 1000} KiB/sec")

    print("Verifying file... ", end="")
    hash_sum = getattr(hashlib, hash_algo)()
    t0 = time.ticks_ms()
    total_size = 0
    with open(path, "rb") as file:
        for i in range(block_count):
            buf = file.read(block_size)
            total_size += len(buf)
            # Update digest
            hash_sum.update(buf)
    dt = time.ticks_diff(time.ticks_ms(), t0)
    print(f"finished: {total_size / 1024 / dt * 1000} KiB/sec; ", end="")

    if digest != hash_sum.digest():
        raise RuntimeError(f"{hash_algo} checksum verify failed")

    print(f"{hash_algo} checksum verified")


if __name__ == "__main__":
    flash_block_test()
    flash_make_filesystem()
    flash_write_verify("test0.bin", size=64 * 1024, block_size=1024)
    flash_write_verify("test1.bin", size=128 * 1024, block_size=1024)
    flash_write_verify("test2.bin", size=64 * 1024, block_size=2048)
    flash_write_verify("test4.bin", size=256 * 1024, block_size=4096)
    flash_write_verify("test4.bin", size=512 * 1024, block_size=16384)



#!/usr/bin/env python3
#
# This file is part of the MicroPython project, http://micropython.org/
#
# The MIT License (MIT)
#
# Copyright (c) 2024 Volodymyr Shymanskyy
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import os
import re
import hashlib
import functools
import pickle

from elftools.elf import elffile
from collections import defaultdict

try:
    from ar import Archive
except:
    Archive = None


class PickleCache:
    def __init__(self, path, prefix=""):
        self.path = path
        self._get_fn = lambda key: os.path.join(path, prefix + key[:24])

    def store(self, key, data):
        os.makedirs(self.path, exist_ok=True)
        # See also https://bford.info/cachedir/
        cachedir_tag_path = os.path.join(self.path, "CACHEDIR.TAG")
        if not os.path.exists(cachedir_tag_path):
            with open(cachedir_tag_path, "w") as f:
                f.write(
                    "Signature: 8a477f597d28d172789f06886806bc55\n"
                    "# This file is a cache directory tag created by MicroPython.\n"
                    "# For information about cache directory tags see https://bford.info/cachedir/\n"
                )
        with open(self._get_fn(key), "wb") as f:
            pickle.dump(data, f)

    def load(self, key):
        with open(self._get_fn(key), "rb") as f:
            return pickle.load(f)


def cached(key, cache):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            cache_key = key(*args, **kwargs)
            try:
                d = cache.load(cache_key)
                if d["key"] != cache_key:
                    raise Exception("Cache key mismatch")
                return d["data"]
            except Exception:
                res = func(*args, **kwargs)
                try:
                    cache.store(
                        cache_key,
                        {
                            "key": cache_key,
                            "data": res,
                        },
                    )
                except Exception:
                    pass
                return res

        return wrapper

    return decorator


class CachedArFile:
    def __init__(self, fn):
        if not Archive:
            raise RuntimeError("Please run 'pip install ar' to link .a files")
        self.fn = fn
        self._archive = Archive(open(fn, "rb"))
        info = self.load_symbols()
        self.objs = info["objs"]
        self.symbols = info["symbols"]

    def open(self, obj):
        return self._archive.open(obj, "rb")

    def _cache_key(self):
        sha = hashlib.sha256()
        with open(self.fn, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                sha.update(chunk)
        # Change this salt if the cache data format changes
        sha.update(bytes.fromhex("00000000000000000000000000000001"))
        return sha.hexdigest()

    @cached(key=_cache_key, cache=PickleCache(path=".mpy_ld_cache", prefix="ar_"))
    def load_symbols(self):
        print("Loading", self.fn)
        objs = defaultdict(lambda: {"def": set(), "undef": set(), "weak": set()})
        symbols = {}
        for entry in self._archive:
            obj_name = entry.name
            elf = elffile.ELFFile(self.open(obj_name))
            symtab = elf.get_section_by_name(".symtab")
            if not symtab:
                continue

            obj = objs[obj_name]

            for symbol in symtab.iter_symbols():
                sym_name = symbol.name
                sym_bind = symbol["st_info"]["bind"]

                if sym_bind in ("STB_GLOBAL", "STB_WEAK"):
                    if symbol.entry["st_shndx"] != "SHN_UNDEF":
                        obj["def"].add(sym_name)
                        symbols[sym_name] = obj_name
                    else:
                        obj["undef"].add(sym_name)

                    if sym_bind == "STB_WEAK":
                        obj["weak"].add(sym_name)

        return {"objs": dict(objs), "symbols": symbols}


def resolve(archives, symbols):
    resolved_objs = []  # Object files needed to resolve symbols
    unresolved_symbols = set()
    provided_symbols = {}  # Which symbol is provided by which object
    symbol_stack = list(symbols)

    # A helper function to handle symbol resolution from a particular object
    def add_obj(archive, symbol):
        obj_name = archive.symbols[symbol]
        obj_info = archive.objs[obj_name]

        obj_tuple = (archive, obj_name)
        if obj_tuple in resolved_objs:
            return  # Already processed this object

        resolved_objs.append(obj_tuple)

        # Add the symbols this object defines
        for defined_symbol in obj_info["def"]:
            if defined_symbol in provided_symbols and not defined_symbol.startswith(
                "__x86.get_pc_thunk."
            ):
                if defined_symbol in obj_info["weak"]:
                    continue
                else:
                    raise RuntimeError(f"Multiple definitions for {defined_symbol}")
            provided_symbols[defined_symbol] = obj_name  # TODO: mark weak if needed

        # Recursively add undefined symbols from this object
        for undef_symbol in obj_info["undef"]:
            if undef_symbol in obj_info["weak"]:
                print(f"Skippping weak dependency: {undef_symbol}")
                continue
            if undef_symbol not in provided_symbols:
                symbol_stack.append(undef_symbol)  # Add undefined symbol to resolve

    while symbol_stack:
        symbol = symbol_stack.pop(0)

        if symbol in provided_symbols:
            continue  # Symbol is already resolved

        found = False
        for archive in archives:
            if symbol in archive.symbols:
                add_obj(archive, symbol)
                found = True
                break

        if not found:
            unresolved_symbols.add(symbol)

    return resolved_objs, list(unresolved_symbols)


def expand_ld_script(fn):
    # This function parses a subset of ld scripts
    # Typically these are just groups of static lib references
    group_pattern = re.compile(r"GROUP\s*\(\s*([^\)]+)\s*\)", re.MULTILINE)
    output_format_pattern = re.compile(r"OUTPUT_FORMAT\s*\(\s*([^\)]+)\s*\)", re.MULTILINE)
    comment_pattern = re.compile(r"/\*.*?\*/", re.MULTILINE | re.DOTALL)

    with open(fn, "r") as f:
        content = f.read()
        content = comment_pattern.sub("", content).strip()

    # Ensure no unrecognized instructions
    leftovers = content
    for pattern in (group_pattern, output_format_pattern):
        leftovers = pattern.sub("", leftovers)
    if leftovers.strip():
        raise ValueError("Invalid instruction found in the ld script:" + leftovers)

    # Extract files from GROUP instructions
    files = []
    for match in group_pattern.findall(content):
        files.extend([file.strip() for file in re.split(r"[,\s]+", match) if file.strip()])

    return files


def load_archive(fn):
    ar_header = b"!<arch>\012"
    with open(fn, "rb") as f:
        is_ar_file = f.read(len(ar_header)) == ar_header
    if is_ar_file:
        return [CachedArFile(fn)]
    else:
        return [CachedArFile(item) for item in expand_ld_script(fn)]



# test constants in assembler


@micropython.asm_rv32
def c1():
    li(a0, 0xFFFFFFFF)
    li(a1, 0xF0000000)
    sub(a0, a0, a1)


print(hex(c1()))


@micropython.asm_rv32
def c2():
    lui(a0, 0x12345)
    li(a1, 0x678)
    add(a0, a0, a1)


print(hex(c2()))


@micropython.asm_rv32
def c3() -> uint:
    lui(a0, 0)
    addi(a0, a0, 0x7FF)


print(hex(c3()))


@micropython.asm_rv32
def c4() -> uint:
    lui(a0, 0)
    addi(a0, a0, -1)


print(hex(c4()))


@micropython.asm_rv32
def c5():
    c_lui(a0, 1)
    c_li(a1, 1)
    c_add(a0, a1)


print(hex(c5()))



# Test uctypes.addressof().

try:
    from sys import maxsize
    import uctypes
except ImportError:
    print("SKIP")
    raise SystemExit

# Test small addresses.
for i in range(8):
    print(uctypes.addressof(uctypes.bytearray_at(1 << i, 8)))

# Test address that is bigger than the greatest small-int but still within the address range.
large_addr = maxsize + 1
print(uctypes.addressof(uctypes.bytearray_at(large_addr, 8)) == large_addr)



# Regression test for DMA for DCache coherency bugs with cache line
# written originally for https://github.com/micropython/micropython/issues/13471

# IMPORTANT: This test requires SPI2 MISO (pin Y8 on Pyboard D) to be connected to GND

import unittest
from machine import SPI

SPI_NUM = 2
BAUDRATE = 5_000_000


class Test(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.spi = SPI(SPI_NUM, baudrate=BAUDRATE)
        cls.skip_slow_test = False

    def test_variable_offset_fixed_length(self):
        buf = bytearray(1024)
        for offs in range(0, len(buf)):
            v = memoryview(buf)[offs : offs + 128]
            self.spi.readinto(v, 0xFF)
            ok = all(b == 0x00 for b in v)
            if not ok:
                print(offs, v.hex())
                self.skip_slow_test = True
            self.assertTrue(ok)

    def test_variable_offset_and_lengths(self):
        # this takes around 30s to run, so skipped if already failing
        if self.skip_slow_test:
            self.skipTest("already failing")

        buf = bytearray(1024)
        for op_len in range(1, 66):
            print(op_len)
            wr = b"\xff" * op_len
            for offs in range(1, len(buf) - op_len - 1):
                # Place some "sentinel" values before and after the DMA buffer
                before = offs & 0xFF
                after = (~offs) & 0xFF
                buf[offs - 1] = before
                buf[offs + op_len] = after
                v = memoryview(buf)[offs : offs + op_len]
                self.spi.write_readinto(wr, v)
                ok = (
                    all(b == 0x00 for b in v)
                    and buf[offs - 1] == before
                    and buf[offs + op_len] == after
                )
                if not ok:
                    print(v.hex())
                    print(hex(op_len), hex(offs), hex(buf[offs - 1]), hex(buf[offs + op_len]))
                self.assertTrue(ok)


if __name__ == "__main__":
    unittest.main()



# MIT license; Copyright (c) 2022 Damien P. George

import struct, sys, os

try:
    from mpy_cross import run as mpy_cross_run
except ImportError:
    mpy_cross_run = None


class VfsRomWriter:
    ROMFS_HEADER = b"\xd2\xcd\x31"

    ROMFS_RECORD_KIND_UNUSED = 0
    ROMFS_RECORD_KIND_PADDING = 1
    ROMFS_RECORD_KIND_DATA_VERBATIM = 2
    ROMFS_RECORD_KIND_DATA_POINTER = 3
    ROMFS_RECORD_KIND_DIRECTORY = 4
    ROMFS_RECORD_KIND_FILE = 5

    def __init__(self):
        self._dir_stack = [(None, bytearray())]

    def _encode_uint(self, value):
        encoded = [value & 0x7F]
        value >>= 7
        while value != 0:
            encoded.insert(0, 0x80 | (value & 0x7F))
            value >>= 7
        return bytes(encoded)

    def _pack(self, kind, payload):
        return self._encode_uint(kind) + self._encode_uint(len(payload)) + payload

    def _extend(self, data):
        buf = self._dir_stack[-1][1]
        buf.extend(data)
        return len(buf)

    def finalise(self):
        _, data = self._dir_stack.pop()
        encoded_kind = VfsRomWriter.ROMFS_HEADER
        encoded_len = self._encode_uint(len(data))
        if (len(encoded_kind) + len(encoded_len) + len(data)) % 2 == 1:
            encoded_len = b"\x80" + encoded_len
        data = encoded_kind + encoded_len + data
        return data

    def opendir(self, dirname):
        self._dir_stack.append((dirname, bytearray()))

    def closedir(self):
        dirname, dirdata = self._dir_stack.pop()
        dirdata = self._encode_uint(len(dirname)) + bytes(dirname, "ascii") + dirdata
        self._extend(self._pack(VfsRomWriter.ROMFS_RECORD_KIND_DIRECTORY, dirdata))

    def mkdata(self, data):
        assert len(self._dir_stack) == 1
        return self._extend(self._pack(VfsRomWriter.ROMFS_RECORD_KIND_DATA_VERBATIM, data)) - len(
            data
        )

    def mkfile(self, filename, filedata):
        filename = bytes(filename, "ascii")
        payload = self._encode_uint(len(filename))
        payload += filename
        if isinstance(filedata, tuple):
            sub_payload = self._encode_uint(filedata[0])
            sub_payload += self._encode_uint(filedata[1])
            payload += self._pack(VfsRomWriter.ROMFS_RECORD_KIND_DATA_POINTER, sub_payload)
        else:
            payload += self._pack(VfsRomWriter.ROMFS_RECORD_KIND_DATA_VERBATIM, filedata)
        self._dir_stack[-1][1].extend(self._pack(VfsRomWriter.ROMFS_RECORD_KIND_FILE, payload))


def copy_recursively(vfs, src_dir, print_prefix, mpy_cross):
    assert src_dir.endswith("/")
    DIR = 1 << 14
    mpy_cross_missed = 0
    dir_contents = sorted(os.listdir(src_dir))
    for name in dir_contents:
        src_name = src_dir + name
        st = os.stat(src_name)

        if name == dir_contents[-1]:
            # Last entry in the directory listing.
            print_entry = "\\--"
            print_recurse = "    "
        else:
            # Not the last entry in the directory listing.
            print_entry = "|--"
            print_recurse = "|   "

        if st[0] & DIR:
            # A directory, enter it and copy its contents recursively.
            print(print_prefix + print_entry, name + "/")
            vfs.opendir(name)
            mpy_cross_missed += copy_recursively(
                vfs, src_name + "/", print_prefix + print_recurse, mpy_cross
            )
            vfs.closedir()
        else:
            # A file.
            did_mpy = False
            name_extra = ""
            if mpy_cross and name.endswith(".py"):
                name_mpy = name[:-3] + ".mpy"
                src_name_mpy = src_dir + name_mpy
                if not os.path.isfile(src_name_mpy):
                    if mpy_cross_run is not None:
                        did_mpy = True
                        proc = mpy_cross_run(src_name)
                        proc.wait()
                    else:
                        mpy_cross_missed += 1
            if did_mpy:
                name_extra = " -> .mpy"
            print(print_prefix + print_entry, name + name_extra)
            if did_mpy:
                name = name_mpy
                src_name = src_name_mpy
            with open(src_name, "rb") as src:
                vfs.mkfile(name, src.read())
            if did_mpy:
                os.remove(src_name_mpy)
    return mpy_cross_missed


def make_romfs(src_dir, *, mpy_cross):
    if not src_dir.endswith("/"):
        src_dir += "/"

    vfs = VfsRomWriter()

    # Build the filesystem recursively.
    print("Building romfs filesystem, source directory: {}".format(src_dir))
    print("/")
    try:
        mpy_cross_missed = copy_recursively(vfs, src_dir, "", mpy_cross)
    except OSError as er:
        print("Error: OSError {}".format(er), file=sys.stderr)
        sys.exit(1)

    if mpy_cross_missed:
        print("Warning: `mpy_cross` module not found, .py files were not precompiled")
        mpy_cross = False

    return vfs.finalise()



try:
    import frozen_asm_thumb as frozen_asm
except ImportError:
    try:
        import frozen_asm_rv32 as frozen_asm
    except ImportError:
        raise ImportError

print(frozen_asm.asm_add(1, 2))
print(frozen_asm.asm_add1(3))
print(frozen_asm.asm_cast_bool(0), frozen_asm.asm_cast_bool(3))
print(frozen_asm.asm_shift_int(4))
print(frozen_asm.asm_shift_uint(4))



print("foobar".endswith("bar"))
print("foobar".endswith("baR"))
print("foobar".endswith("bar1"))
print("foobar".endswith("foobar"))
print("foobar".endswith(""))
print("foobar".endswith("foobarbaz"))

print("foobar".endswith("bar", 3))
print("foobar".endswith("bar", 4))
print("foobar".endswith("foo", 0, 3))
print("foobar".endswith("foo", 0, 4))
print("foobar".endswith("foo", 1, 3))
print("foobar".endswith("foo", 1, 3))
print("foobar".endswith("oo", 1, 3))
print("foobar".endswith("o", 2, 3))
print("foobar".endswith("o", 3, 3))
print("foobar".endswith("o", 4, 3))

print("foobar".endswith("bar", None, None))
print("foobar".endswith("bar", None, 3))
print("foobar".endswith("bar", 3, None))
print("foobar".endswith("bar", 2, None))
print("foobar".endswith("foo", None, 3))

print("foobar".endswith(("bar", "foo")))
print("foobar".endswith(("foo", "bar")))
print("foobar".endswith(("foo", "bar1")))
print("foobar".endswith(("bar", )))
print("foobar".endswith(("foo", )))

try:
    "foobar".endswith(1)
except TypeError:
    print("TypeError")
